#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
seedsplit: Seed split classes and methods for the MMGen suite
"""

from .color import yellow
from .util import msg, die
from .objmethods import MMGenObject, HiliteStr, InitErrors
from .obj import ImmutableAttr, MMGenPWIDString, MMGenIdx, get_obj, IndexedDict
from .seed import Seed, SeedBase
from .subseed import SubSeedList, SubSeedIdx, SubSeed
from .crypto import Crypto

class SeedShareIdx(MMGenIdx):
	max_val = 1024

class SeedShareCount(SeedShareIdx):
	min_val = 2

class MasterShareIdx(MMGenIdx):
	max_val = 1024

class SeedSplitSpecifier(HiliteStr, InitErrors, MMGenObject):
	color = 'red'
	def __new__(cls, s):
		if isinstance(s, cls):
			return s
		try:
			arr = s.split(':')
			assert len(arr) in (2, 3), 'cannot be parsed'
			a, b, c = arr if len(arr) == 3 else ['default'] + arr
			me = str.__new__(cls, s)
			me.id = SeedSplitIDString(a)
			me.idx = SeedShareIdx(b)
			me.count = SeedShareCount(c)
			assert me.idx <= me.count, 'share index greater than share count'
			return me
		except Exception as e:
			return cls.init_fail(e, s)

def is_seed_split_specifier(s):
	return get_obj(SeedSplitSpecifier, s=s, silent=True, return_bool=True)

class SeedSplitIDString(MMGenPWIDString):
	desc = 'seed split ID string'

class SeedShareList(SubSeedList):
	have_short = False
	split_type = 'N-of-N'

	count  = ImmutableAttr(SeedShareCount)
	id_str = ImmutableAttr(SeedSplitIDString)

	def __init__(self, parent_seed, count, *, id_str=None, master_idx=None, debug_last_share=False):
		self.member_type = SeedShare
		self.parent_seed = parent_seed
		self.id_str = id_str or 'default'
		self.count = count
		self.len = 2 # placeholder, always overridden

		def make_master_share():
			for nonce in range(SeedShare.max_nonce+1):
				ms = SeedShareMaster(self, master_idx, nonce)
				if ms.sid == parent_seed.sid:
					if parent_seed.cfg.debug_subseed:
						msg(f'master_share seed ID collision with parent seed, incrementing nonce to {nonce+1}')
				else:
					return ms
			die('SubSeedNonceRangeExceeded', 'nonce range exceeded')

		def last_share_debug(last_share):
			if not debug_last_share:
				return False
			sid_len = self.debug_last_share_sid_len
			lsid = last_share.sid[:sid_len]
			psid = parent_seed.sid[:sid_len]
			ssids = [d[:sid_len] for d in self.data['long'].keys]
			return (lsid in ssids or lsid == psid)

		self.master_share = make_master_share() if master_idx else None

		for nonce in range(SeedShare.max_nonce+1):
			self.nonce_start = nonce
			self.data = {'long': IndexedDict(), 'short': IndexedDict()} # 'short' is required as a placeholder
			if self.master_share:
				self.data['long'][self.master_share.sid] = (1, self.master_share.nonce)
			self._generate(count-1)
			self.last_share = ls = SeedShareLast(self)
			if last_share_debug(ls) or ls.sid in self.data['long'] or ls.sid == parent_seed.sid:
				# collision: throw out entire split list and redo with new start nonce
				if parent_seed.cfg.debug_subseed:
					self._collision_debug_msg(
						ls.sid, count, nonce, nonce_desc='nonce_start', debug_last_share=debug_last_share)
			else:
				self.data['long'][ls.sid] = (count, nonce)
				break
		else:
			die('SubSeedNonceRangeExceeded', 'nonce range exceeded')

		if parent_seed.cfg.debug_subseed:
			A = parent_seed.data
			B = self.join().data
			assert A == B, f'Data mismatch!\noriginal seed: {A!r}\nrejoined seed: {B!r}'

	def get_share_by_idx(self, idx, *, base_seed=False):
		if idx < 1 or idx > self.count:
			die('RangeError', f'{idx}: share index out of range')
		elif idx == self.count:
			return self.last_share
		elif self.master_share and idx == 1:
			return self.master_share if base_seed else self.master_share.derived_seed
		else:
			ss_idx = SubSeedIdx(str(idx) + 'L')
			return self.get_subseed_by_ss_idx(ss_idx)

	def get_share_by_seed_id(self, sid, *, base_seed=False):
		if sid == self.data['long'].key(self.count-1):
			return self.last_share
		elif self.master_share and sid == self.data['long'].key(0):
			return self.master_share if base_seed else self.master_share.derived_seed
		else:
			return self.get_subseed_by_seed_id(sid)

	def join(self):
		return Seed.join_shares(
			self.parent_seed.cfg,
			[self.get_share_by_idx(i+1) for i in range(len(self))])

	def format(self):
		assert self.split_type == 'N-of-N'
		fs1 = '    {}\n'
		fs2 = '{i:>5}: {}\n'
		mfs1, mfs2, midx, msid = ('', '', '', '')
		if self.master_share:
			mfs1, mfs2 = (' with master share #{} ({})', ' (master share #{})')
			midx, msid = (self.master_share.idx, self.master_share.sid)

		hdr  = '    {} {} ({} bits)\n'.format('Seed:', self.parent_seed.sid.hl(), self.parent_seed.bitlen)
		hdr += '    {} {c}-of-{c} (XOR){m}\n'.format('Split Type:', c=self.count, m=mfs1.format(midx, msid))
		hdr += '    {} {}\n\n'.format('ID String:', self.id_str.hl())
		hdr += fs1.format('Shares')
		hdr += fs1.format('------')

		sl = self.data['long'].keys
		body1 = fs2.format(sl[0]+mfs2.format(midx), i=1)
		body = (fs2.format(sl[n], i=n+1) for n in range(1, len(self)))

		return hdr + body1 + ''.join(body)

class SeedShareBase(MMGenObject):

	@property
	def fn_stem(self):
		pl = self.parent_list
		msdata = f'_with_master{pl.master_share.idx}' if pl.master_share else ''
		return '{}-{}-{}of{}{}[{}]'.format(
			pl.parent_seed.sid,
			pl.id_str,
			self.idx,
			pl.count,
			msdata,
			self.sid)

	@property
	def desc(self):
		return self.get_desc()

	def get_desc(self, *, ui=False):
		pl = self.parent_list
		mss = f', with master share #{pl.master_share.idx}' if pl.master_share else ''
		if ui:
			m   = (yellow("(share {} of {} of ")
					+ pl.parent_seed.sid.hl()
					+ yellow(', split id ')
					+ pl.id_str.hl2(encl='‘’')
					+ yellow('{})'))
		else:
			m = "share {} of {} of " + pl.parent_seed.sid + ", split id '" + pl.id_str + "'{}"
		return m.format(self.idx, pl.count, mss)

class SeedShare(SeedShareBase, SubSeed):

	@staticmethod
	def make_subseed_bin(parent_list, idx:int, nonce:int, length:str):
		seed = parent_list.parent_seed
		assert parent_list.have_short is False
		assert length == 'long'
		# field maximums: id_str: none (256 chars), count: 65535 (1024), idx: 65535 (1024), nonce: 65535 (1000)
		scramble_key = (
			f'{parent_list.split_type}:{parent_list.id_str}:'.encode() +
			parent_list.count.to_bytes(2, 'big') +
			idx.to_bytes(2, 'big') +
			nonce.to_bytes(2, 'big')
		)
		if parent_list.master_share:
			scramble_key += (
				b':master:' +
				parent_list.master_share.idx.to_bytes(2, 'big')
			)
		return Crypto(parent_list.parent_seed.cfg).scramble_seed(seed.data, scramble_key)[:seed.byte_len]

class SeedShareLast(SeedShareBase, SeedBase):

	idx = ImmutableAttr(SeedShareIdx)
	nonce = 0

	def __init__(self, parent_list):
		self.idx = parent_list.count
		self.parent_list = parent_list
		SeedBase.__init__(
			self,
			parent_list.parent_seed.cfg,
			seed_bin=self.make_subseed_bin(parent_list))

	@staticmethod
	def make_subseed_bin(parent_list):
		seed_list = (parent_list.get_share_by_idx(i+1) for i in range(len(parent_list)))
		seed = parent_list.parent_seed

		ret = int(seed.data.hex(), 16)
		for ss in seed_list:
			ret ^= int(ss.data.hex(), 16)

		return ret.to_bytes(seed.byte_len, 'big')

class SeedShareMaster(SeedBase, SeedShareBase):

	idx   = ImmutableAttr(MasterShareIdx)
	nonce = ImmutableAttr(int, typeconv=False)

	def __init__(self, parent_list, idx, nonce):
		self.idx = idx
		self.nonce = nonce
		self.parent_list = parent_list
		self.cfg = parent_list.parent_seed.cfg

		SeedBase.__init__(self, self.cfg, seed_bin=self.make_base_seed_bin())

		self.derived_seed = SeedBase(
			self.cfg,
			seed_bin = self.make_derived_seed_bin(parent_list.id_str, parent_list.count))

	@property
	def fn_stem(self):
		return '{}-MASTER{}[{}]'.format(self.parent_list.parent_seed.sid, self.idx, self.sid)

	def make_base_seed_bin(self):
		seed = self.parent_list.parent_seed
		# field maximums: idx: 65535 (1024)
		scramble_key = b'master_share:' + self.idx.to_bytes(2, 'big') + self.nonce.to_bytes(2, 'big')
		return Crypto(self.cfg).scramble_seed(seed.data, scramble_key)[:seed.byte_len]

	# Don't bother with avoiding seed ID collision here, as sid of derived seed is not used
	# by user as an identifier
	def make_derived_seed_bin(self, id_str, count):
		# field maximums: id_str: none (256 chars), count: 65535 (1024)
		scramble_key = id_str.encode() + b':' + count.to_bytes(2, 'big')
		return Crypto(self.cfg).scramble_seed(self.data, scramble_key)[:self.byte_len]

	def get_desc(self, *, ui=False):
		psid = self.parent_list.parent_seed.sid
		mss = f'master share #{self.idx} of '
		return yellow('(' + mss) + psid.hl() + yellow(')') if ui else mss + psid

class SeedShareMasterJoining(SeedShareMaster):

	id_str = ImmutableAttr(SeedSplitIDString)
	count  = ImmutableAttr(SeedShareCount)

	def __init__(self, cfg, idx, base_seed, id_str, count):

		SeedBase.__init__(self, cfg, seed_bin=base_seed.data)

		self.cfg = cfg
		self.id_str = id_str or 'default'
		self.count = count
		self.derived_seed = SeedBase(
			cfg,
			seed_bin = self.make_derived_seed_bin(self.id_str, self.count))

def join_shares(
		cfg,
		seed_list,
		*,
		master_idx = None,
		id_str     = None):

	if not hasattr(seed_list, '__next__'): # seed_list can be iterator or iterable
		seed_list = iter(seed_list)

	class d:
		byte_len, ret, count = None, 0, 0

	def add_share(ss):
		if d.byte_len:
			assert ss.byte_len == d.byte_len, f'Seed length mismatch! {ss.byte_len} != {d.byte_len}'
		else:
			d.byte_len = ss.byte_len
		d.ret ^= int(ss.data.hex(), 16)
		d.count += 1

	if master_idx:
		master_share = next(seed_list)

	for ss in seed_list:
		add_share(ss)

	if master_idx:
		add_share(SeedShareMasterJoining(cfg, master_idx, master_share, id_str, d.count+1).derived_seed)

	SeedShareCount(d.count) # check that d.count is in valid range

	return Seed(cfg, seed_bin=d.ret.to_bytes(d.byte_len, 'big'))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.base: Bitcoin base transaction class
"""

from collections import namedtuple

from ....tx.base import Base as TxBase
from ....obj import MMGenList, HexStr, ListItemAttr
from ....util import msg, make_chksum_6, die, pp_fmt

from .op_return_data import OpReturnData

def data2scriptPubKey(data):
	return '6a' + '{:02x}'.format(len(data)) + data.hex() # OP_RETURN data

def addr2scriptPubKey(proto, addr):

	def decode_addr(proto, addr):
		ap = proto.decode_addr(addr)
		assert ap, f'coin address {addr!r} could not be parsed'
		return ap.bytes.hex()

	return {
		'p2pkh': '76a914' + decode_addr(proto, addr) + '88ac',
		'p2sh':  'a914' + decode_addr(proto, addr) + '87',
		'bech32': proto.witness_vernum_hex + '14' + decode_addr(proto, addr)
	}[addr.addr_fmt]

def decodeScriptPubKey(proto, s):
	# src/wallet/rpc/addresses.cpp:
	#   types: nonstandard, pubkey, pubkeyhash, scripthash, multisig, nulldata, witness_v0_keyhash
	ret = namedtuple('decoded_scriptPubKey', ['type', 'addr_fmt', 'addr', 'data'])

	if len(s) == 50 and s[:6] == '76a914' and s[-4:] == '88ac':
		return ret('pubkeyhash', 'p2pkh', proto.pubhash2addr(bytes.fromhex(s[6:-4]), 'p2pkh'), None)

	elif len(s) == 46 and s[:4] == 'a914' and s[-2:] == '87':
		return ret('scripthash', 'p2sh', proto.pubhash2addr(bytes.fromhex(s[4:-2]), 'p2sh'), None)

	elif len(s) == 44 and s[:4] == proto.witness_vernum_hex + '14':
		return ret('witness_v0_keyhash', 'bech32', proto.pubhash2bech32addr(bytes.fromhex(s[4:])), None)

	elif s[:2] == '6a': # OP_RETURN
		# range 1-80 == hex 2-160, plus 4 for opcode byte + push byte
		if 6 <= len(s) <= (proto.max_op_return_data_len * 2) + 6: # 2-160 -> 6-166
			return ret('nulldata', None, None, s[4:]) # return data in hex format
		else:
			raise ValueError('{}: OP_RETURN data bytes length not in range 1-{}'.format(
					len(s[4:]) // 2,
					proto.max_op_return_data_len))

	else:
		raise NotImplementedError(f'Unrecognized scriptPubKey ({s})')

def DeserializeTX(proto, txhex):
	"""
	Parse a serialized Bitcoin transaction
	For checking purposes, additionally reconstructs the serialized TX without signature
	"""

	def bytes2int(bytes_le):
		return int(bytes_le[::-1].hex(), 16)

	def bytes2coin_amt(bytes_le):
		return proto.coin_amt(bytes2int(bytes_le), from_unit='satoshi')

	def bshift(n, *, skip=False, sub_null=False):
		nonlocal idx, raw_tx
		ret = tx[idx:idx+n]
		idx += n
		if sub_null:
			raw_tx += b'\x00'
		elif not skip:
			raw_tx += ret
		return ret

	# https://bitcoin.org/en/developer-reference#compactsize-unsigned-integers
	# For example, the number 515 is encoded as 0xfd0302.
	def readVInt(*, skip=False):
		nonlocal idx, raw_tx
		s = tx[idx]
		idx += 1
		if not skip:
			raw_tx.append(s)

		vbytes_len = 1 if s < 0xfd else 2 if s == 0xfd else 4 if s == 0xfe else 8

		if vbytes_len == 1:
			return s
		else:
			vbytes = tx[idx:idx+vbytes_len]
			idx += vbytes_len
			if not skip:
				raw_tx += vbytes
			return int(vbytes[::-1].hex(), 16)

	def make_txid(tx_bytes):
		from hashlib import sha256
		return sha256(sha256(tx_bytes).digest()).digest()[::-1].hex()

	tx = bytes.fromhex(txhex)
	raw_tx = bytearray()
	idx = 0

	d = {'version': bytes2int(bshift(4))}

	if d['version'] > 0x7fffffff: # version is signed integer
		die(3, f"{d['version']}: transaction version greater than maximum allowed value (int32_t)!")

	has_witness = tx[idx] == 0
	if has_witness:
		u = bshift(2, skip=True).hex()
		if u != '0001':
			die('IllegalWitnessFlagValue', f'{u!r}: Illegal value for flag in transaction!')

	d['num_txins'] = readVInt()

	d['txins'] = MMGenList([{
		'txid':      bshift(32)[::-1].hex(),
		'vout':      bytes2int(bshift(4)),
		'scriptSig': bshift(readVInt(skip=True), sub_null=True).hex(),
		'nSeq':      bshift(4)[::-1].hex()
	} for i in range(d['num_txins'])])

	d['num_txouts'] = readVInt()

	d['txouts'] = MMGenList([{
		'amt':          bytes2coin_amt(bshift(8)),
		'scriptPubKey': bshift(readVInt()).hex()
	} for i in range(d['num_txouts'])])

	for o in d['txouts']:
		o.update(decodeScriptPubKey(proto, o['scriptPubKey'])._asdict())

	if has_witness:
		# https://github.com/bitcoin/bips/blob/master/bip-0141.mediawiki
		# A non-witness program (defined hereinafter) txin MUST be associated with an empty
		# witness field, represented by a 0x00.

		d['txid'] = make_txid(tx[:4] + tx[6:idx] + tx[-4:])
		d['witness_size'] = len(tx) - idx + 2 - 4 # add len(marker+flag), subtract len(locktime)

		for txin in d['txins']:
			if tx[idx] == 0:
				bshift(1, skip=True)
				continue
			txin['witness'] = [
				bshift(readVInt(skip=True), skip=True).hex() for item in range(readVInt(skip=True))]
	else:
		d['txid'] = make_txid(tx)
		d['witness_size'] = 0

	if len(tx) - idx != 4:
		die('TxHexParseError', 'TX hex has invalid length: {} extra bytes'.format(len(tx)-idx-4))

	d['locktime'] = bytes2int(bshift(4))
	d['unsigned_hex'] = raw_tx.hex()

	return namedtuple('deserialized_tx', list(d.keys()))(**d)

class Base(TxBase):
	rel_fee_desc = 'satoshis per byte'
	rel_fee_disp = 'sat/byte'
	_deserialized = None

	class Output(TxBase.Output): # output contains either addr or data, but not both
		data = ListItemAttr(OpReturnData, include_proto=True) # type None in parent cls

	class InputList(TxBase.InputList):

		# Lexicographical Indexing of Transaction Inputs and Outputs
		# https://github.com/bitcoin/bips/blob/master/bip-0069.mediawiki
		def sort_bip69(self):
			def sort_func(a):
				return (
					bytes.fromhex(a.txid)
					+ int.to_bytes(a.vout, 4, 'big'))
			self.sort(key=sort_func)

	class OutputList(TxBase.OutputList):

		def sort_bip69(self):
			def sort_func(a):
				return (
					int.to_bytes(a.amt.to_unit('satoshi'), 8, 'big')
					+ bytes.fromhex(
						addr2scriptPubKey(self.parent.proto, a.addr) if a.addr else
						data2scriptPubKey(a.data)))
			self.sort(key=sort_func)

	def has_segwit_inputs(self):
		return any(i.mmtype in ('S', 'B') for i in self.inputs)

	def has_segwit_outputs(self):
		return any(o.mmtype in ('S', 'B') for o in self.outputs)

	# https://bitcoin.stackexchange.com/questions/1195/how-to-calculate-transaction-size-before-sending
	# 180: uncompressed, 148: compressed
	def estimate_size_old(self):
		if not self.inputs or not self.outputs:
			return None
		return len(self.inputs)*180 + len(self.outputs)*34 + 10

	# https://bitcoincore.org/en/segwit_wallet_dev/
	# vsize: 3 times of the size with original serialization, plus the size with new
	# serialization, divide the result by 4 and round up to the next integer.

	# TODO: results differ slightly from actual transaction size
	def estimate_size(self):
		if not self.inputs or not self.outputs:
			return None

		sig_size = 72 # sig in DER format
		pubkey_size_uncompressed = 65
		pubkey_size_compressed = 33

		def get_inputs_size():
			# txid vout [scriptSig size (vInt)] scriptSig (<sig> <pubkey>) nSeq
			isize_common = 32 + 4 + 1 + 4 # txid vout [scriptSig size] nSeq = 41
			input_size = {
				'L': isize_common + sig_size + pubkey_size_uncompressed, # = 180
				'C': isize_common + sig_size + pubkey_size_compressed,   # = 148
				'S': isize_common + 23,                                  # = 64
				'B': isize_common + 0                                    # = 41
			}
			ret = sum(input_size[i.mmtype] for i in self.inputs if i.mmtype)

			# We have no way of knowing whether a non-MMGen P2PKH addr is compressed or uncompressed
			# until we see the key, so assume compressed for fee-estimation purposes. If fee estimate
			# is off by more than 5%, sign() aborts and user is instructed to use --vsize-adj option.
			return ret + sum(input_size['C'] for i in self.inputs if not i.mmtype)

		def get_outputs_size():
			# output bytes:
			#   8 (amt) + scriptlen_byte + script_bytes
			#   script_bytes:
			#     ADDR: p2pkh: 25, p2sh: 23, bech32: 22
			#     DATA: opcode_byte ('6a') + push_byte + nulldata_bytes
			return sum(
				{'p2pkh':34, 'p2sh':32, 'bech32':31}[o.addr.addr_fmt] if o.addr else
				(11 + len(o.data)) if o.data else
				# guess value if o.addr is missing (probably a vault address):
				34 if self.proto.coin == 'BCH' else
				31
					for o in self.outputs)

		# https://github.com/bitcoin/bips/blob/master/bip-0141.mediawiki
		# The witness is a serialization of all witness data of the transaction. Each txin is
		# associated with a witness field. A witness field starts with a var_int to indicate the
		# number of stack items for the txin. It is followed by stack items, with each item starts
		# with a var_int to indicate the length. Witness data is NOT script.

		# A non-witness program txin MUST be associated with an empty witness field, represented
		# by a 0x00. If all txins are not witness program, a transaction's wtxid is equal to its txid.
		def get_witness_size():
			if not self.has_segwit_inputs():
				return 0
			wf_size = 1 + 1 + sig_size + 1 + pubkey_size_compressed # vInt vInt sig vInt pubkey = 108
			return sum((1, wf_size)[i.mmtype in ('S', 'B')] for i in self.inputs)

		isize = get_inputs_size()
		osize = get_outputs_size()
		wsize = get_witness_size()

		# TODO: compute real varInt sizes instead of assuming 1 byte
		# Serialization:
		#   old:     [nVersion]              [vInt][txins][vInt][txouts]         [nLockTime]
		old_size =   4                     + 1   + isize + 1  + osize          + 4
		#   marker = 0x00, flag = 0x01
		#   new:     [nVersion][marker][flag][vInt][txins][vInt][txouts][witness][nLockTime]
		new_size =   4       + 1     + 1   + 1   + isize + 1  + osize + wsize  + 4 if wsize else old_size

		ret = (old_size * 3 + new_size) // 4

		self.cfg._util.dmsg(
			'\nData from estimate_size():\n' +
			f'  inputs size: {isize}, outputs size: {osize}, witness size: {wsize}\n' +
			f'  size: {new_size}, vsize: {ret}, old_size: {old_size}')

		return int(ret * (self.cfg.vsize_adj or 1))

	# convert absolute CoinAmt fee to sat/byte for display using estimated size
	def fee_abs2rel(self, abs_fee, *, to_unit='satoshi'):
		return str(int(
			abs_fee /
			getattr(self.proto.coin_amt, to_unit) /
			self.estimate_size()))

	@property
	def data_output(self):
		res = self.data_outputs
		if len(res) > 1:
			raise ValueError(f'{res}: too many data outputs in transaction (only one allowed)')
		return res[0] if len(res) == 1 else None

	@data_output.setter
	def data_output(self, val):
		dbool = [bool(o.data) for o in self.outputs]
		if dbool.count(True) != 1:
			raise ValueError('more or less than one data output in transaction!')
		self.outputs[dbool.index(True)] = val

	@property
	def data_outputs(self):
		return [o for o in self.outputs if o.data]

	@property
	def nondata_outputs(self):
		return [o for o in self.outputs if not o.data]

	@property
	def deserialized(self):
		if not self._deserialized:
			self._deserialized = DeserializeTX(self.proto, self.serialized)
		return self._deserialized

	def update_serialized(self, data):
		self.serialized = HexStr(data)
		self._deserialized = None
		self.check_serialized_integrity()

	def check_serialized_integrity(self):
		"""
		Check that a malicious, compromised or malfunctioning coin daemon hasn't produced bad
		serialized tx data.

		Does not check witness data.

		Perform this check every time a serialized tx is received from the coin daemon or read
		from a transaction file.
		"""

		def do_error(errmsg):
			die('TxHexMismatch', errmsg+'\n'+hdr)

		def check_equal(desc, hexio, mmio):
			if mmio != hexio:
				msg('\nMMGen {d}:\n{m}\nSerialized {d}:\n{h}'.format(
					d = desc,
					m = pp_fmt(mmio),
					h = pp_fmt(hexio)))
				do_error(
					f'{desc.capitalize()} in serialized transaction data from coin daemon ' +
					'do not match those in MMGen transaction!')

		hdr = 'A malicious or malfunctioning coin daemon or other program may have altered your data!'

		dtx = self.deserialized

		if dtx.locktime != int(self.locktime or 0):
			do_error(
				f'Transaction hex nLockTime ({dtx.locktime}) ' +
				f'does not match MMGen transaction nLockTime ({self.locktime})')

		check_equal(
			'sequence numbers',
			[i['nSeq'] for i in dtx.txins],
			['{:08x}'.format(i.sequence or self.proto.max_int) for i in self.inputs])

		check_equal(
			'inputs',
			sorted((i['txid'], i['vout']) for i in dtx.txins),
			sorted((i.txid, i.vout) for i in self.inputs))

		check_equal(
			'outputs',
			sorted((o['addr'] or o['data'], o['amt']) for o in dtx.txouts),
			sorted((o.addr or o.data.hex(), o.amt) for o in self.outputs))

		if str(self.txid) != make_chksum_6(bytes.fromhex(dtx.unsigned_hex)).upper():
			do_error(f'MMGen TxID ({self.txid}) does not match serialized transaction data!')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.online: Bitcoin online signed transaction class
"""

from ....tx import online as TxBase
from ....util import msg, die
from ....color import orange
from .signed import Signed

class OnlineSigned(Signed, TxBase.OnlineSigned):

	async def send_checks(self):

		self.check_correct_chain()

		if not self.cfg.bogus_send:
			if self.has_segwit_outputs() and not self.rpc.info('segwit_is_active'):
				die(2, 'Transaction has Segwit outputs, but this blockchain does not support Segwit'
						+ ' at the current height')

		if self.fee > self.proto.max_tx_fee:
			die(2, 'Transaction fee ({}) greater than {} max_tx_fee ({} {})!'.format(
				self.fee,
				self.proto.name,
				self.proto.max_tx_fee,
				self.proto.coin))

		await self.status.display()

	async def test_sendable(self, txhex):

		await self.send_checks()

		res = await self.rpc.call('testmempoolaccept', (txhex,))
		ret = res[0]

		if ret['allowed']:
			from ....obj import CoinTxID
			msg('TxID: {}'.format(CoinTxID(ret['txid']).hl()))
			return True
		else:
			msg(ret['reject-reason'])
			return False

	async def send_with_node(self, txhex):
		try:
			return await self.rpc.call('sendrawtransaction', txhex)
		except Exception as e:
			errmsg = str(e)
			nl = '\n'
			if errmsg.count('Signature must use SIGHASH_FORKID'):
				m = (
					'The Aug. 1 2017 UAHF has activated on this chain.\n'
					'Re-run the script with the --coin=bch option.')
			elif errmsg.count('Illegal use of SIGHASH_FORKID'):
				m  = (
					'The Aug. 1 2017 UAHF is not yet active on this chain.\n'
					'Re-run the script without the --coin=bch option.')
			elif errmsg.count('non-final'):
				m = "Transaction with nLockTime {!r} can’t be included in this block!".format(
					self.info.strfmt_locktime(self.get_serialized_locktime()))
			else:
				m, nl = ('', '')
			msg(orange('\n'+errmsg))
			die(2, f'{m}{nl}Send of MMGen transaction {self.txid} failed')

	async def post_network_send(self, coin_txid):
		return True

class Sent(TxBase.Sent, OnlineSigned):
	pass

class AutomountOnlineSigned(TxBase.AutomountOnlineSigned, OnlineSigned):
	pass

class AutomountSent(TxBase.AutomountSent, AutomountOnlineSigned):
	pass



#!/usr/bin/env python3

import sys, os
pn = os.path.abspath(os.path.dirname(sys.argv[0]))
parpar = os.path.dirname(os.path.dirname(pn))
os.chdir(parpar)
sys.path[0] = os.curdir

from mmgen.cfg import Config, gc
from mmgen.color import yellow, blue, cyan, set_vt100
from mmgen.util import msg, ymsg, gmsg, fmt, fmt_list, die

commands = [
	'start',
	'get_terminal_size',
	'color',
	'license',
	'line_input',
	'urand',
	'txview',
	'get_char_one',
	'get_char_one_raw',
]
if sys.platform in ('linux', 'darwin'):
	commands.extend([
		'get_char',
		'get_char_immed_chars',
		'get_char_raw',
	])
elif sys.platform == 'win32':
	commands.extend([
		'get_char_one_char_immed_chars',
	])

opts_data = {
	'text': {
		'desc': 'Interactively test MMGen terminal functionality',
		'usage': 'command',
		'options': """
-h, --help     Print this help message
""",
	'notes': f"""
available commands for platform {sys.platform!r}:
{fmt_list(commands, fmt='col', indent='    ')}
"""
	}
}

cfg = Config(opts_data=opts_data)

from mmgen.term import get_char_raw, get_terminal_size, get_term
from mmgen.ui import line_input, keypress_confirm, do_license_msg
import mmgen.term as term_mod

def cmsg(m):
	msg('\n'+cyan(m))

def confirm(m):
	if not keypress_confirm(cfg, m):
		if keypress_confirm(cfg, 'Are you sure you want to exit test?'):
			die(1, 'Exiting test at user request')
		else:
			msg('Continuing...')

def tt_start():
	m = fmt("""
		We will now test MMGen Wallet’s terminal capabilities.
		This is a non-automated test and requires user interaction.
		Continue?
	""")
	confirm(m.strip())

def tt_get_terminal_size():
	cmsg('Testing get_terminal_size():')
	msg('X' * get_terminal_size().width)
	confirm('Do the X’s exactly fill the width of the screen?')

def tt_color():
	cmsg('Testing color:')
	confirm(blue('THIS TEXT') + ' should be blue.  Is it?')

def tt_license():
	cmsg('Testing do_license_msg() with pager')
	ymsg('Press "w" to test the pager, then "c" to continue')
	do_license_msg(cfg)

def tt_line_input():
	set_vt100()
	cmsg('Testing line_input():')
	msg(fmt("""
		At the Ready? prompt type and hold down "y".
		Then Enter some text, followed by held-down ENTER.
		The held-down "y" and ENTER keys should be blocked, not affecting the output
		on screen or entered text.
	"""))
	get_char_raw('Ready? ', num_bytes=1)
	reply = line_input(cfg, '\nEnter text: ')
	confirm(f'Did you enter the text {reply!r}?')

def _tt_get_char(raw=False, one_char=False, immed_chars=''):
	funcname = ('get_char', 'get_char_raw')[raw]
	fs = fmt("""
		Press some keys in quick succession.
		{}{}
		{}
		When you’re finished, use Ctrl-C to exit.
		""").strip()
	m1 = (
		'You should experience a delay with quickly repeated entry.',
		'Your entry should be repeated back to you immediately.'
	)[raw]
	m2 = (
		'',
		f'\nThe characters {immed_chars!r} will be repeated immediately, the others with delay.'
	)[bool(immed_chars)]
	m3 = 'The F1-F12 keys will be ' + (
		'blocked entirely.'
			if one_char and not raw else
		"echoed AS A SINGLE character '\\x1b'."
			if one_char else
		'echoed as a FULL CONTROL SEQUENCE.'
	)
	if sys.platform == 'win32':
		if raw:
			m3 = 'The Escape and F1-F12 keys will be returned as two-character strings.'
		else:
			m3 = 'The Escape and F1-F12 keys will be returned as single characters.'
	kwargs = {}
	if one_char:
		kwargs.update({'num_bytes':1})
	if immed_chars:
		kwargs.update({'immed_chars':immed_chars})

	cmsg('Testing {}({}):'.format(
		funcname,
		','.join(f'{a}={b!r}' for a, b in kwargs.items())
	))
	msg(fs.format(m1, yellow(m2), yellow(m3)))

	try:
		while True:
			ret = getattr(term_mod, funcname)('Enter a letter: ', **kwargs)
			msg(f'You typed {ret!r}')
	except KeyboardInterrupt:
		msg('\nDone')

def tt_urand():
	cmsg('Testing _get_random_data_from_user():')
	from mmgen.crypto import Crypto
	ret = Crypto(cfg)._get_random_data_from_user(uchars=10, desc='data').decode()
	msg(f'USER ENTROPY (user input + keystroke timings):\n\n{fmt(ret, "  ")}')
	times = ret.splitlines()[1:]
	avg_prec = sum(len(t.split('.')[1]) for t in times) // len(times)
	if avg_prec < gc.min_time_precision:
		ymsg(f'WARNING: Avg. time precision of only {avg_prec} decimal points.  User entropy quality is degraded!')
	else:
		msg(f'Average time precision: {avg_prec} decimal points - OK')
	line_input(cfg, 'Press ENTER to continue: ')

def tt_txview():
	cmsg('Testing tx.info.view_with_prompt() (try each viewing option)')
	from mmgen.tx import UnsignedTX
	fn = 'test/ref/0B8D5A[15.31789,14,tl=1320969600].rawtx'
	tx = UnsignedTX(cfg=cfg, filename=fn, quiet_open=True)
	while True:
		tx.info.view_with_prompt('View data for transaction?', pause=False)
		set_vt100()
		if not keypress_confirm(cfg, 'Continue testing transaction view?', default_yes=True):
			break

def tt_get_char_one():
	_tt_get_char(one_char=True)

def tt_get_char_one_raw():
	_tt_get_char(one_char=True, raw=True)

def tt_get_char():
	_tt_get_char(one_char=False)

def tt_get_char_immed_chars():
	_tt_get_char(one_char=False, immed_chars='asdf')

def tt_get_char_raw():
	_tt_get_char(one_char=False, raw=True)

def tt_get_char_one_char_immed_chars():
	_tt_get_char(one_char=True, immed_chars='asdf')

get_term().register_cleanup()

if cfg._args:
	locals()['tt_'+cfg._args[0]]()
else:
	for command in commands:
		locals()['tt_'+command]()

gmsg('\nTest completed')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
passwdlist: Password list class for the MMGen suite
"""

from collections import namedtuple

from .util import ymsg, is_int, die
from .obj import ImmutableAttr, ListItemAttr, MMGenPWIDString, TwComment
from .key import PrivKey
from .addr import MMGenPasswordType, AddrIdx, AddrListID
from .addrlist import (
	AddrListChksum,
	AddrListIDStr,
	AddrListEntryBase,
	AddrList,
)

class PasswordListEntry(AddrListEntryBase):
	passwd  = ListItemAttr(str, typeconv=False) # TODO: create Password type
	idx     = ImmutableAttr(AddrIdx)
	comment = ListItemAttr(TwComment, reassign_ok=True)
	sec     = ListItemAttr(PrivKey, include_proto=True)

class PasswordList(AddrList):
	entry_type  = PasswordListEntry
	main_attr   = 'passwd'
	desc        = 'password'
	gen_desc    = 'password'
	gen_desc_pl = 's'
	gen_addrs   = False
	gen_keys    = False
	gen_passwds = True
	pw_len      = None
	dfl_pw_fmt  = 'b58'
	pwinfo      = namedtuple('passwd_info', ['min_len', 'max_len', 'dfl_len', 'valid_lens', 'desc', 'chk_func'])
	pw_info     = {
		# 32**25 < 2**128 < 32**26
		'b32':     pwinfo(10, 42 , 24, None,         'base32 password',           'baseconv.is_b32_str'),
		# 58**21 < 2**128 < 58**22
		'b58':     pwinfo(8,  36 , 20, None,         'base58 password',           'baseconv.is_b58_str'),
		'bip39':   pwinfo(12, 24 , 24, [12, 18, 24], 'BIP39 mnemonic',            'bip39.is_bip39_mnemonic'),
		'xmrseed': pwinfo(25, 25,  25, [25],         'Monero new-style mnemonic', 'xmrseed.is_xmrseed'),
		'hex':     pwinfo(32, 64 , 64, [32, 48, 64], 'hexadecimal password',      'util.is_hex_str'),
	}
	chksum_rec_f = lambda foo, e: (str(e.idx), e.passwd)

	feature_warn_fs = 'WARNING: {!r} is a potentially dangerous feature.  Use at your own risk!'
	hex2bip39 = False

	def __init__(
			self,
			cfg,
			proto,
			*,
			infile          = None,
			seed            = None,
			pw_idxs         = None,
			pw_id_str       = None,
			pw_len          = None,
			pw_fmt          = None,
			chk_params_only = False,
			skip_chksum_msg = False):

		self.cfg = cfg
		self.proto = proto # proto is ignored

		if not cfg.debug_addrlist:
			self.dmsg_sc = self.noop

		if infile:
			self.infile = infile
			# sets self.pw_id_str, self.pw_fmt, self.pw_len, self.chk_func:
			self.data = self.file.parse_file(infile)
		else:
			if not chk_params_only:
				for k in (seed, pw_idxs):
					assert k
			self.pw_id_str = MMGenPWIDString(pw_id_str)
			self.set_pw_fmt(pw_fmt)
			self.set_pw_len(pw_len)
			if chk_params_only:
				return
			if self.hex2bip39:
				ymsg(self.feature_warn_fs.format(pw_fmt))
			self.set_pw_len_vs_seed_len(seed) # sets self.bip39, self.xmrseed, self.xmrproto self.baseconv
			self.al_id = AddrListID(sid=seed.sid, mmtype=MMGenPasswordType(self.proto, 'P'))
			self.data = self.generate(seed, pw_idxs)

		self.num_addrs = len(self.data)
		self.fmt_data = ''
		self.chksum = AddrListChksum(self)

		fs = f'{self.al_id.sid}-{self.pw_id_str}-{self.pw_fmt_disp}-{self.pw_len}[{{}}]'
		self.id_str = AddrListIDStr(self, fmt_str=fs)

		if not skip_chksum_msg:
			self.do_chksum_msg(record=not infile)

	def set_pw_fmt(self, pw_fmt):
		if pw_fmt == 'hex2bip39':
			self.hex2bip39 = True
			self.pw_fmt = 'bip39'
			self.pw_fmt_disp = 'hex2bip39'
		else:
			self.pw_fmt = pw_fmt
			self.pw_fmt_disp = pw_fmt
		if self.pw_fmt not in self.pw_info:
			die('InvalidPasswdFormat',
				f'{self.pw_fmt!r}: invalid password format.  Valid formats: {", ".join(self.pw_info)}')

	def chk_pw_len(self):
		assert self.pw_len, 'pw_len must be set'
		pw_len = self.pw_len
		fs = '{l}: invalid user-requested length for {b} ({c}{m})'
		d = self.pw_info[self.pw_fmt]
		if d.valid_lens:
			if pw_len not in d.valid_lens:
				die(2, fs.format(l=pw_len, b=d.desc, c='not one of ', m=d.valid_lens))
		elif pw_len > d.max_len:
			die(2, fs.format(l=pw_len, b=d.desc, c='>', m=d.max_len))
		elif pw_len < d.min_len:
			die(2, fs.format(l=pw_len, b=d.desc, c='<', m=d.min_len))

	def set_pw_len(self, pw_len):
		d = self.pw_info[self.pw_fmt]

		if pw_len is None:
			self.pw_len = d.dfl_len
			return

		if not is_int(pw_len):
			die(2, f'{pw_len!r}: invalid user-requested password length (not an integer)')
		self.pw_len = int(pw_len)
		self.chk_pw_len()

	def set_pw_len_vs_seed_len(self, seed):
		pf = self.pw_fmt
		if pf == 'hex':
			pw_bytes = self.pw_len // 2
			good_pw_len = seed.byte_len * 2
		elif pf == 'bip39':
			from .bip39 import bip39
			self.bip39 = bip39()
			pw_bytes = bip39.nwords2seedlen(self.pw_len, in_bytes=True)
			good_pw_len = bip39.seedlen2nwords(seed.byte_len, in_bytes=True)
		elif pf == 'xmrseed':
			from .xmrseed import xmrseed
			from .protocol import init_proto
			self.xmrseed = xmrseed()
			self.xmrproto = init_proto(self.cfg, 'xmr')
			pw_bytes = xmrseed().seedlen_map_rev[self.pw_len]
			try:
				good_pw_len = xmrseed().seedlen_map[seed.byte_len]
			except:
				die(1, f'{seed.byte_len*8}: unsupported seed length for Monero new-style mnemonic')
		elif pf in ('b32', 'b58'):
			pw_int = (32 if pf == 'b32' else 58) ** self.pw_len
			pw_bytes = pw_int.bit_length() // 8
			from .baseconv import baseconv
			self.baseconv = baseconv(self.pw_fmt)
			good_pw_len = len(baseconv(pf).frombytes(b'\xff'*seed.byte_len))
		else:
			raise NotImplementedError(f'{pf!r}: unknown password format')

		if pw_bytes > seed.byte_len:
			die(1,
				f'Cannot generate passwords with more entropy than underlying seed! ({len(seed.data)*8} bits)\n' +
				(f'Re-run the command with --passwd-len={good_pw_len}' if pf in ('bip39', 'hex') else
				'Re-run the command, specifying a password length of {} or less')
			)

		if pf in ('bip39', 'hex') and pw_bytes < seed.byte_len:
			from .ui import keypress_confirm
			keypress_confirm(
				self.cfg,
				f'WARNING: requested {self.pw_info[pf].desc} length has less entropy ' +
				'than underlying seed!\nIs this what you want?',
				default_yes = True,
				do_exit = True)

	def gen_passwd(self, secbytes):
		assert self.pw_fmt in self.pw_info
		if self.pw_fmt == 'hex':
			# take most significant part
			return secbytes.hex()[:self.pw_len]
		elif self.pw_fmt == 'bip39':
			pw_len_bytes = self.bip39.nwords2seedlen(self.pw_len, in_bytes=True)
			# take most significant part
			return ' '.join(self.bip39.fromhex(secbytes[:pw_len_bytes].hex()))
		elif self.pw_fmt == 'xmrseed':
			pw_len_bytes = self.xmrseed.seedlen_map_rev[self.pw_len]
			bytes_preproc = self.xmrproto.preprocess_key(
				secbytes[:pw_len_bytes], # take most significant part
				None)
			return ' '.join(self.xmrseed.frombytes(bytes_preproc))
		else:
			# take least significant part
			return self.baseconv.frombytes(
				secbytes,
				pad = self.pw_len,
				tostr = True)[-self.pw_len:]

	def check_format(self, pw):
		if not self.chk_func(pw):
			raise ValueError(f'Password is not valid {self.pw_info[self.pw_fmt].desc} data')
		pwlen = len(pw.split()) if self.pw_fmt in ('bip39', 'xmrseed') else len(pw)
		if pwlen != self.pw_len:
			raise ValueError(f'Password has incorrect length ({pwlen} != {self.pw_len})')
		return True

	def scramble_seed(self, seed):
		# Changing either pw_fmt or pw_len will cause a different, unrelated
		# set of passwords to be generated: this is what we want.
		# NB: In original implementation, pw_id_str was 'baseN', not 'bN'
		scramble_key = f'{self.pw_fmt}:{self.pw_len}:{self.pw_id_str}'

		if self.hex2bip39:
			pwlen = self.bip39.nwords2seedlen(self.pw_len, in_hex=True)
			scramble_key = f'hex:{pwlen}:{self.pw_id_str}'

		self.dmsg_sc('str', scramble_key)
		from .crypto import Crypto
		return Crypto(self.cfg).scramble_seed(seed, scramble_key.encode())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.spec: Monero wallet ops for the MMGen Suite
"""

import re

from ...util import die
from ...addr import CoinAddr

from .. import uarg_info

class OpMixinSpec:

	def create_addr_data(self):
		m = re.fullmatch(uarg_info[self.spec_id].pat, self.uargs.spec, re.ASCII)
		if not m:
			fs = "{!r}: invalid {!r} arg: for {} operation, it must have format {!r}"
			die(1, fs.format(self.uargs.spec, self.spec_id, self.name, uarg_info[self.spec_id].annot))

		def gen():
			for i, k in self.spec_key:
				if m[i] is None:
					setattr(self, k, None)
				else:
					idx = int(m[i])
					try:
						res = self.kal.entry(idx)
					except:
						die(1, f'Supplied key-address file does not contain address {self.kal.al_id.sid}:{idx}')
					else:
						setattr(self, k, res)
						yield res

		self.addr_data = list(gen())
		self.account = None if m[2] is None else int(m[2])

		def strip_quotes(s):
			if s and s[0] in ("'", '"'):
				if s[-1] != s[0] or len(s) < 2:
					die(1, f'{s!r}: unbalanced quotes in label string!')
				return s[1:-1]
			else:
				return s # None or empty string

		if self.name in ('sweep', 'sweep_all'):
			self.dest_acct = None if m[4] is None else int(m[4])
		elif self.name == 'transfer':
			self.dest_addr = CoinAddr(self.proto, m[3])
			self.amount = self.proto.coin_amt(m[4])
		elif self.name == 'new':
			self.label = strip_quotes(m[3])
		elif self.name == 'label':
			self.address_idx = int(m[3])
			self.label = strip_quotes(m[4])



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.file.tx: Monero transaction file class for the MMGen Suite
"""

import time
from collections import namedtuple
from pathlib import Path

from ...obj import CoinTxID, Int
from ...color import red, yellow, blue, cyan, pink, orange, purple
from ...util import die, fmt, make_timestr, list_gen
from ...seed import SeedID
from ...protocol import init_proto
from ...addr import CoinAddr
from ...tx.util import get_autosign_obj

from ..include import XMRWalletAddrSpec
from . import MoneroMMGenFile

class MoneroMMGenTX:

	class Base(MoneroMMGenFile):

		data_label = 'MoneroMMGenTX'

		# both base_chksum and full_chksum are used to make the filename stem, so we must not include
		# fields that change when TX is signed and submitted (e.g. ‘sign_time’, ‘submit_time’)
		base_chksum_fields = {
			'op',
			'create_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'amount' }
		full_chksum_fields = {
			'op',
			'create_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'amount',
			'fee',
			'blob' }
		oneline_fs = '{a:7} {b:8} {c:19} {d:13} {e:9} {f:6} {x:2} {g:6} {h:17} {j}'
		oneline_fixed_cols_w = 96 # width of all columns except the last (coin address)
		chksum_nchars = 6
		xmrwallet_tx_data = namedtuple('xmrwallet_tx_data', [
			'op',
			'create_time',
			'sign_time',
			'submit_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'dest_address',
			'txid',
			'amount',
			'priority',
			'fee',
			'blob',
			'metadata',
			'unsigned_txset',
			'signed_txset',
			'complete',
		])

		def __init__(self):
			self.name = type(self).__name__

		@property
		def src_wallet_idx(self):
			return int(self.data.source.split(':')[0])

		def get_info_oneline(self, *, indent='', addr_w=None):
			d = self.data
			return self.oneline_fs.format(
					a = yellow(d.network),
					b = d.seed_id.hl(),
					c = make_timestr(d.submit_time if d.submit_time is not None else d.create_time),
					d = orange(self.file_id),
					e = purple(d.op.ljust(9)),
					f = red('{}:{}'.format(d.source.wallet, d.source.account).ljust(6)),
					g = red('{}:{}'.format(d.dest.wallet, d.dest.account).ljust(6)) if d.dest else cyan('ext   '),
					h = d.amount.fmt(4, color=True, prec=12),
					j = d.dest_address.fmt(0, addr_w, color=True) if addr_w else d.dest_address.hl(0),
					x = '->'
				)

		def get_info(self, *, indent='', addr_w=None):
			d = self.data
			pmt_id = d.dest_address.parsed.payment_id
			fs = '\n'.join(list_gen(
				['Info for transaction {a} [Seed ID: {b}. Network: {c}]:'],
				['  TxID:      {d}'],
				['  Created:   {e:19} [{f}]'],
				['  Signed:    {g:19} [{h}]', d.sign_time],
				['  Submitted: {s:19} [{t}]', d.submit_time],
				['  Type:      {i}{S}'],
				['  From:      wallet {j}, account {k}'],
				['  To:        wallet {x}, account {y}, address {z}', d.dest],
				['  Amount:    {m} XMR'],
				['  Priority:  {F}', d.priority],
				['  Fee:       {n} XMR'],
				['  Dest:      {o}'],
				['  Size:      {Z} bytes', d.signed_txset],
				['  Payment ID: {P}', pmt_id],
			))

			from ...util2 import format_elapsed_hr
			from ..ops import addr_width
			from .. import tx_priorities
			return fmt(fs, strip_char='\t', indent=indent).format(
					a = orange(self.file_id),
					b = d.seed_id.hl(),
					c = yellow(d.network.upper()),
					d = d.txid.hl(),
					e = make_timestr(d.create_time),
					f = format_elapsed_hr(d.create_time),
					g = make_timestr(d.sign_time) if d.sign_time else None,
					h = format_elapsed_hr(d.sign_time) if d.sign_time else None,
					i = blue(d.op),
					j = d.source.wallet.hl(),
					k = red(f'#{d.source.account}'),
					m = d.amount.hl(),
					F = (Int(d.priority).hl() + f' [{tx_priorities[d.priority]}]') if d.priority else None,
					n = d.fee.hl(),
					o = d.dest_address.hl(0) if self.cfg.full_address
						else d.dest_address.fmt(0, addr_width, color=True),
					P = pink(pmt_id.hex()) if pmt_id else None,
					s = make_timestr(d.submit_time) if d.submit_time else None,
					S = pink(f" [cold signed{', submitted' if d.complete else ''}]") if d.signed_txset else '',
					t = format_elapsed_hr(d.submit_time) if d.submit_time else None,
					x = d.dest.wallet.hl() if d.dest else None,
					y = red(f'#{d.dest.account}') if d.dest else None,
					z = red(f'#{d.dest.account_address}') if d.dest else None,
					Z = Int(len(d.signed_txset) // 2).hl() if d.signed_txset else None,
				)

		@property
		def file_id(self):
			return (self.base_chksum + ('-' + self.full_chksum if self.full_chksum else '')).upper()

		def write(self, *, delete_metadata=False, ask_write=True, ask_overwrite=True):
			dict_data = self.data._asdict()
			if delete_metadata:
				dict_data['metadata'] = None

			fn = '{a}-XMR[{b!s}]{c}.{d}'.format(
				a = self.file_id,
				b = self.data.amount,
				c = '' if self.data.network == 'mainnet' else f'.{self.data.network}',
				d = self.ext
			)

			if self.cfg.autosign:
				fn = get_autosign_obj(self.cfg).xmr_tx_dir / fn

			from ...fileutil import write_data_to_file
			write_data_to_file(
				cfg                   = self.cfg,
				outfile               = str(fn),
				data                  = self.make_wrapped_data(dict_data),
				desc                  = self.desc,
				ask_write             = ask_write,
				ask_write_default_yes = not ask_write,
				ask_overwrite         = ask_overwrite,
				ignore_opt_outdir     = self.cfg.autosign)

	class New(Base):

		def __init__(self, *args, **kwargs):

			super().__init__()

			assert not args, 'Non-keyword args not permitted'

			if '_in_tx' in kwargs:
				in_data = kwargs.pop('_in_tx').data._asdict()
				in_data.update(kwargs)
			else:
				in_data = kwargs

			d = namedtuple('monero_tx_in_data_tuple', in_data)(**in_data)
			self.cfg = d.cfg

			proto = init_proto(self.cfg, 'xmr', network=d.network, need_amt=True)

			now = int(time.time())

			self.data = self.xmrwallet_tx_data(
				op             = d.op,
				create_time    = now if self.name in ('NewSigned', 'NewUnsigned') else getattr(d, 'create_time', None),
				sign_time      = now if self.name in ('NewSigned', 'NewColdSigned') else getattr(d, 'sign_time', None),
				submit_time    = now if self.name == 'NewSubmitted' else None,
				network        = d.network,
				seed_id        = SeedID(sid=d.seed_id),
				source         = XMRWalletAddrSpec(d.source),
				dest           = None if d.dest is None else XMRWalletAddrSpec(d.dest),
				dest_address   = CoinAddr(proto, d.dest_address),
				txid           = CoinTxID(d.txid),
				amount         = d.amount,
				priority       = self.cfg.priority if self.name in ('NewSigned', 'NewUnsigned') else d.priority,
				fee            = d.fee,
				blob           = d.blob,
				metadata       = d.metadata,
				unsigned_txset = d.unsigned_txset,
				signed_txset   = getattr(d, 'signed_txset', None),
				complete       = self.name in ('NewSigned', 'NewSubmitted'),
			)

	class NewUnsigned(New):
		desc = 'unsigned transaction'
		ext = 'rawtx'
		signed = False

	class NewSigned(New):
		desc = 'signed transaction'
		ext = 'sigtx'
		signed = True

	class NewColdSigned(NewSigned):
		pass

	class NewSubmitted(NewColdSigned):
		desc = 'submitted transaction'
		ext = 'subtx'

	class Completed(Base):
		desc = 'transaction'
		forbidden_fields = ()

		def __init__(self, cfg, fn):

			super().__init__()

			self.cfg = cfg
			self.fn = Path(fn)

			try:
				d_wrap = self.extract_data_from_file(cfg, fn)
			except Exception as e:
				die('MoneroMMGenTXFileParseError', f'{type(e).__name__}: {e}\nCould not load transaction file')

			if 'unsigned_txset' in d_wrap['data']: # post-autosign
				self.full_chksum_fields &= set(d_wrap['data']) # allow for added chksum fields in future
			else:
				self.full_chksum_fields = set(d_wrap['data']) - {'metadata'}

			for key in self.xmrwallet_tx_data._fields: # backwards compat: fill in missing fields
				if not key in d_wrap['data']:
					d_wrap['data'][key] = None

			d = self.xmrwallet_tx_data(**d_wrap['data'])

			if self.name not in ('View', 'Completed'):
				assert fn.name.endswith('.'+self.ext), 'TX file {fn} has incorrect extension (not {self.ext!r})'
				assert getattr(d, self.req_field), f'{self.name} TX missing required field {self.req_field!r}'
				assert bool(d.sign_time) == self.signed, '{a} has {b}sign time!'.format(
					a = self.desc,
					b = 'no ' if self.signed else'')
				for f in self.forbidden_fields:
					assert not getattr(d, f), f'{self.name} TX mismatch: contains forbidden field {f!r}'

			proto = init_proto(cfg, 'xmr', network=d.network, need_amt=True)

			self.data = self.xmrwallet_tx_data(
				op             = d.op,
				create_time    = d.create_time,
				sign_time      = d.sign_time,
				submit_time    = d.submit_time,
				network        = d.network,
				seed_id        = SeedID(sid=d.seed_id),
				source         = XMRWalletAddrSpec(d.source),
				dest           = None if d.dest is None else XMRWalletAddrSpec(d.dest),
				dest_address   = CoinAddr(proto, d.dest_address),
				txid           = CoinTxID(d.txid),
				amount         = proto.coin_amt(d.amount),
				priority       = d.priority,
				fee            = proto.coin_amt(d.fee),
				blob           = d.blob,
				metadata       = d.metadata,
				unsigned_txset = d.unsigned_txset,
				signed_txset   = d.signed_txset,
				complete       = d.complete,
			)

			self.check_checksums(d_wrap)

	class Unsigned(Completed):
		desc = 'unsigned transaction'
		ext = 'rawtx'
		signed = False
		req_field = 'unsigned_txset'
		forbidden_fields = ('signed_txset',)

	class Signed(Completed):
		desc = 'signed transaction'
		ext = 'sigtx'
		signed = True
		req_field = 'blob'
		forbidden_fields = ('signed_txset', 'unsigned_txset')

	class ColdSigned(Signed):
		req_field = 'signed_txset'
		forbidden_fields = ()

	class Submitted(ColdSigned):
		desc = 'submitted transaction'
		ext = 'subtx'
		silent_load = True

	class View(Completed):
		silent_load = True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
cmdtest_d.include.group_mgr: Command group manager for the MMGen Wallet cmdtest suite
"""

import sys, os, time

from mmgen.color import yellow, green, cyan
from mmgen.util import Msg, die

from .cfg import cfgs, cmd_groups_dfl, cmd_groups_extra

class CmdGroupMgr:

	dpy_data = None

	cmd_groups = cmd_groups_dfl.copy()
	cmd_groups.update(cmd_groups_extra)

	cfg_attrs = (
		'seed_len',
		'seed_id',
		'wpasswd',
		'kapasswd',
		'segwit',
		'hash_preset',
		'bw_filename',
		'bw_params',
		'ref_bw_seed_id',
		'addr_idx_list',
		'pass_idx_list')

	def __init__(self, cfg):
		self.cfg = cfg
		self.network_id = cfg._proto.coin.lower() + ('_tn' if cfg._proto.testnet else '')
		self.name = type(self).__name__

	def create_cmd_group(self, cls, sg_name=None):

		cmd_group_in = dict(cls.cmd_group_in)

		if sg_name and 'subgroup.' + sg_name not in cmd_group_in:
			die(1, f'{sg_name!r}: no such subgroup in test group {cls.__name__}')

		def add_entries(key, add_deps=True, added_subgroups=[]):

			if add_deps:
				for dep in cmd_group_in['subgroup.'+key]:
					yield from add_entries(dep)

			assert isinstance(cls.cmd_subgroups[key][0], str), f'header for subgroup {key!r} missing!'

			if not key in added_subgroups:
				yield from cls.cmd_subgroups[key][1:]
				added_subgroups.append(key)

		def gen():
			for name, data in cls.cmd_group_in:
				if name.startswith('subgroup.'):
					sg_key = name.removeprefix('subgroup.')
					if sg_name in (None, sg_key):
						yield from add_entries(
								sg_key,
								add_deps = sg_name and not self.cfg.skipping_deps,
								added_subgroups = [sg_name] if self.cfg.deps_only else [])
					if self.cfg.deps_only and sg_key == sg_name:
						return
				else:
					yield (name, data)

		return tuple(gen())

	def load_mod(self, gname, modname=None):
		clsname, kwargs = self.cmd_groups[gname]
		if modname is None and 'modname' in kwargs:
			modname = kwargs['modname']
		import importlib
		modpath = f'test.cmdtest_d.{modname or gname}'
		return getattr(importlib.import_module(modpath), clsname)

	def create_group(self, gname, sg_name, full_data=False, modname=None, is3seed=False, add_dpy=False):
		"""
		Initializes the list 'cmd_list' and dict 'dpy_data' from module's cmd_group data.
		Alternatively, if called with 'add_dpy=True', updates 'dpy_data' from module data
		without touching 'cmd_list'
		"""

		cls = self.load_mod(gname, modname)
		cdata = []

		def get_shared_deps(cmdname, tmpdir_idx):
			"""
			shared_deps are "implied" dependencies for all cmds in cmd_group that don't appear in
			the cmd_group data or cmds' argument lists.  Supported only for 3seed tests at present.
			"""
			if not hasattr(cls, 'shared_deps'):
				return []

			return [k for k, v in cfgs[str(tmpdir_idx)]['dep_generators'].items()
						if k in cls.shared_deps and v != cmdname]

		if not 'cmd_group' in cls.__dict__ and hasattr(cls, 'cmd_group_in'):
			cls.cmd_group = self.create_cmd_group(cls, sg_name)

		for a, b in cls.cmd_group:
			if is3seed:
				for n, (i, j) in enumerate(zip(cls.tmpdir_nums, (128, 192, 256))):
					k = f'{a}_{n+1}'
					if hasattr(cls, 'skip_cmds') and k in cls.skip_cmds:
						continue
					sdeps = get_shared_deps(k, i)
					if isinstance(b, str):
						cdata.append((k, (i, f'{b} ({j}-bit)', [[[]+sdeps, i]])))
					else:
						cdata.append((k, (i, f'{b[1]} ({j}-bit)', [[b[0]+sdeps, i]])))
			elif full_data:
				cdata.append((a, b))
			else:
				cdata.append((a, (cls.tmpdir_nums[0], b, [[[], cls.tmpdir_nums[0]]])))

		if add_dpy:
			self.dpy_data.update(dict(cdata))
		else:
			self.cmd_list = tuple(e[0] for e in cdata)
			self.dpy_data = dict(cdata)

		cls.full_data = full_data or is3seed

		if not cls.full_data:
			cls.tmpdir_num = cls.tmpdir_nums[0]
			for k, v in cfgs[str(cls.tmpdir_num)].items():
				setattr(cls, k, v)

		return cls

	def gm_init_group(self, cfg, trunner, gname, sg_name, spawn_prog):
		kwargs = self.cmd_groups[gname][1]
		cls = self.create_group(gname, sg_name, **kwargs)
		cls.group_name = gname
		return cls(cfg, trunner, cfgs, spawn_prog)

	def get_cls_by_gname(self, gname):
		return self.load_mod(gname, self.cmd_groups[gname][1].get('modname'))

	def list_cmd_groups(self):
		ginfo = []
		for gname in self.cmd_groups:
			ginfo.append((gname, self.get_cls_by_gname(gname)))

		if self.cfg.list_current_cmd_groups:
			exclude = (self.cfg.exclude_groups or '').split(',')
			ginfo = [g for g in ginfo
						if self.network_id in g[1].networks
							and not g[0] in exclude
							and g[0] in tuple(self.cmd_groups_dfl) + tuple(self.cfg._args)]
			desc = 'CONFIGURED'
		else:
			desc = 'AVAILABLE'

		def gen_output():
			yield green(f'{desc} COMMAND GROUPS AND SUBGROUPS:')
			yield ''
			for name, cls in ginfo:
				if not cls.is_helper:
					yield '  {} - {}'.format(
						yellow(name.ljust(13)),
						(cls.__doc__.strip() if cls.__doc__ else cls.__name__))
					if 'cmd_subgroups' in cls.__dict__:
						subgroups = {k:v for k, v in cls.cmd_subgroups.items() if not k.startswith('_')}
						max_w = max(len(k) for k in subgroups)
						for k, v in subgroups.items():
							yield '    + {} · {}'.format(cyan(k.ljust(max_w+1)), v[0])

		from mmgen.ui import do_pager
		do_pager('\n'.join(gen_output()))

		Msg('\n' + ' '.join(e[0] for e in ginfo))

	def find_cmd_in_groups(self, cmd, group=None):
		"""
		Search for a test command in specified group or all configured command groups
		and return it as a string.  Loads modules but alters no global variables.
		"""
		if group:
			if not group in [e[0] for e in self.cmd_groups]:
				die(1, f'{group!r}: unrecognized group')
			groups = [self.cmd_groups[group]]
		else:
			groups = self.cmd_groups

		for gname in groups:
			cls = self.get_cls_by_gname(gname)

			if not hasattr(cls, 'cmd_group'):
				cls.cmd_group = self.create_cmd_group(cls)

			if cmd in cls.cmd_group:             # first search the class
				return gname

			if cmd in dir(cls(self.cfg, None, None, None)):  # then a throwaway instance
				return gname # cmd might exist in more than one group - we'll go with the first

		return None



#!/usr/bin/env python3

"""
test.modtest_d.bip_hd: bip_hd unit test for the MMGen suite
"""

from mmgen.color import gray, pink, blue
from mmgen.util import fmt
from mmgen.bip_hd import Bip32ExtendedKey, BipHDConfig, BipHDNode, MasterNode, get_chain_params

from ..include.common import cfg, vmsg

# Source: BIP-32
vectors_bip32 = [
{
	'seed': '000102030405060708090a0b0c0d0e0f',
	"m": {
		'xpub': 'xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8',
		'xprv': 'xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi',
	},
	"m/0'": {
		'xpub': 'xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw',
		'xprv': 'xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7',
	},
	"m/0'/1": {
		'xpub': 'xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ',
		'xprv': 'xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYPxLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs',
	},
	"m/0'/1/2'": {
		'xpub': 'xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5',
		'xprv': 'xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3ryjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM',
	},
	"m/0'/1/2'/2": {
		'xpub': 'xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6ZLRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV',
		'xprv': 'xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334',
	},
	"m/0'/1/2'/2/1000000000": {
		'xpub': 'xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy',
		'xprv': 'xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76',
	},
}, {
	'seed': 'fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542',
	'm': {
		'xpub': 'xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDMSgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB',
		'xprv': 'xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U',
	},
	"m/0": {
		'xpub': 'xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERfvrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH',
		'xprv': 'xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2yJD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt',
	},
	"m/0/2147483647'": {
		'xpub': 'xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDBrQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a',
		'xprv': 'xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwCd6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9',
	},
	"m/0/2147483647'/1": {
		'xpub': 'xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoNJxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon',
		'xprv': 'xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef',
	},
	"m/0/2147483647'/1/2147483646'": {
		'xpub': 'xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL',
		'xprv': 'xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39njGVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc',
	},
	"m/0/2147483647'/1/2147483646'/2": {
		'xpub': 'xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGsApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt',
		'xprv': 'xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq38EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j',
	},
}, {
	'comment': 'These vectors test for the retention of leading zeros. See bitpay/bitcore-lib#47 and iancoleman/bip39#58 for more information.',
	'seed': '4b381541583be4423346c643850da4b320e46a87ae3d2a4e6da11eba819cd4acba45d239319ac14f863b8d5ab5a0d0c64d2e8a1e7d1457df2e5a3c51c73235be',
	'm': {
		'xpub': 'xpub661MyMwAqRbcEZVB4dScxMAdx6d4nFc9nvyvH3v4gJL378CSRZiYmhRoP7mBy6gSPSCYk6SzXPTf3ND1cZAceL7SfJ1Z3GC8vBgp2epUt13',
		'xprv': 'xprv9s21ZrQH143K25QhxbucbDDuQ4naNntJRi4KUfWT7xo4EKsHt2QJDu7KXp1A3u7Bi1j8ph3EGsZ9Xvz9dGuVrtHHs7pXeTzjuxBrCmmhgC6',
	},
	"m/0'": {
		'xpub': 'xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y',
		'xprv': 'xprv9uPDJpEQgRQfDcW7BkF7eTya6RPxXeJCqCJGHuCJ4GiRVLzkTXBAJMu2qaMWPrS7AANYqdq6vcBcBUdJCVVFceUvJFjaPdGZ2y9WACViL4L',
	},
}, {
	'comment': 'These vectors test for the retention of leading zeros. See btcsuite/btcutil#172 for more information.',
	'seed': '3ddd5602285899a946114506157c7997e5444528f3003f6134712147db19b678',
	"m": {
		'xpub': 'xpub661MyMwAqRbcGczjuMoRm6dXaLDEhW1u34gKenbeYqAix21mdUKJyuyu5F1rzYGVxyL6tmgBUAEPrEz92mBXjByMRiJdba9wpnN37RLLAXa',
		'xprv': 'xprv9s21ZrQH143K48vGoLGRPxgo2JNkJ3J3fqkirQC2zVdk5Dgd5w14S7fRDyHH4dWNHUgkvsvNDCkvAwcSHNAQwhwgNMgZhLtQC63zxwhQmRv',
	},
	"m/0'": {
		'xpub': 'xpub69AUMk3qDBi3uW1sXgjCmVjJ2G6WQoYSnNHyzkmdCHEhSZ4tBok37xfFEqHd2AddP56Tqp4o56AePAgCjYdvpW2PU2jbUPFKsav5ut6Ch1m',
		'xprv': 'xprv9vB7xEWwNp9kh1wQRfCCQMnZUEG21LpbR9NPCNN1dwhiZkjjeGRnaALmPXCX7SgjFTiCTT6bXes17boXtjq3xLpcDjzEuGLQBM5ohqkao9G',
	},
	"m/0'/1'": {
		'xpub': 'xpub6BJA1jSqiukeaesWfxe6sNK9CCGaujFFSJLomWHprUL9DePQ4JDkM5d88n49sMGJxrhpjazuXYWdMf17C9T5XnxkopaeS7jGk1GyyVziaMt',
		'xprv': 'xprv9xJocDuwtYCMNAo3Zw76WENQeAS6WGXQ55RCy7tDJ8oALr4FWkuVoHJeHVAcAqiZLE7Je3vZJHxspZdFHfnBEjHqU5hG1Jaj32dVoS6XLT1',
	},
}]

# Source: BIP-32
# These vectors test that invalid extended keys are recognized as invalid.
vectors_bip32_invalid = [
	('xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6LBpB85b3D2yc8sfvZU521AAwdZafEz7mnzBBsz4wKY5fTtTQBm', 'pubkey version / prvkey mismatch'),
	('xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGTQQD3dC4H2D5GBj7vWvSQaaBv5cxi9gafk7NF3pnBju6dwKvH', 'prvkey version / pubkey mismatch'),
	('xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Txnt3siSujt9RCVYsx4qHZGc62TG4McvMGcAUjeuwZdduYEvFn', 'invalid pubkey prefix 04'),
	('xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGpWnsj83BHtEy5Zt8CcDr1UiRXuWCmTQLxEK9vbz5gPstX92JQ', 'invalid prvkey prefix 04'),
	('xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6N8ZMMXctdiCjxTNq964yKkwrkBJJwpzZS4HS2fxvyYUA4q2Xe4', 'invalid pubkey prefix 01'),
	('xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD9y5gkZ6Eq3Rjuahrv17fEQ3Qen6J', 'invalid prvkey prefix 01'),
	('xprv9s2SPatNQ9Vc6GTbVMFPFo7jsaZySyzk7L8n2uqKXJen3KUmvQNTuLh3fhZMBoG3G4ZW1N2kZuHEPY53qmbZzCHshoQnNf4GvELZfqTUrcv', 'zero depth with non-zero parent fingerprint'),
	('xpub661no6RGEX3uJkY4bNnPcw4URcQTrSibUZ4NqJEw5eBkv7ovTwgiT91XX27VbEXGENhYRCf7hyEbWrR3FewATdCEebj6znwMfQkhRYHRLpJ', 'zero depth with non-zero parent fingerprint'),
	('xprv9s21ZrQH4r4TsiLvyLXqM9P7k1K3EYhA1kkD6xuquB5i39AU8KF42acDyL3qsDbU9NmZn6MsGSUYZEsuoePmjzsB3eFKSUEh3Gu1N3cqVUN', 'zero depth with non-zero index'),
	('xpub661MyMwAuDcm6CRQ5N4qiHKrJ39Xe1R1NyfouMKTTWcguwVcfrZJaNvhpebzGerh7gucBvzEQWRugZDuDXjNDRmXzSZe4c7mnTK97pTvGS8', 'zero depth with non-zero index'),
	('DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHGMQzT7ayAmfo4z3gY5KfbrZWZ6St24UVf2Qgo6oujFktLHdHY4', 'unknown extended key version'),
	('DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHPmHJiEDXkTiJTVV9rHEBUem2mwVbbNfvT2MTcAqj3nesx8uBf9', 'unknown extended key version'),
	('xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzF93Y5wvzdUayhgkkFoicQZcP3y52uPPxFnfoLZB21Teqt1VvEHx', 'private key 0 not in 1..n-1'),
	('xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD5SDKr24z3aiUvKr9bJpdrcLg1y3G', 'private key n not in 1..n-1'),
	('xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Q5JXayek4PRsn35jii4veMimro1xefsM58PgBMrvdYre8QyULY', 'invalid pubkey 02000000...07'),
	('xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHL', 'invalid checksum'),
]

# Source: bip_utils
vectors_derive = {
	'bech32': {
		0: 'bc1qwg77fxw0tkmc3h58tcnnpegxk7mp3h6ly44d3n',
		1: 'bc1q6g79y6kwpkufevv2njacvnqnsdxmen68jyvjde',
		2: 'bc1qknujpwlxc9e9e6avz50q5k90p552xy8g3qjd8u',
	}
}

# Source: bip_utils
vectors_addrfmt = {
	'pub': {
		'compressed': 'xpub6GJknXsmpFcEubJsddGacncHhyY5Bk9zMQKC8pC97vBVAphrchYxuoJsAqfZW2uEMfPr6umSPRrhuaA7zeuExkwuAWiUcKcXjSf437VMLwR',
		'segwit': 'ypub6aNved9dRKbMRjLfMbGPoXKgNN1tr86qi8WoBxSVr4mHX9fdzUxJFtEH63QZZxArk4f2fwFZQUQ7FRkNiBarTLu2Y69SRxzn68WysngXPrp',
		'bech32': 'zpub6urFg31yVogtpf3Y7aV3CLuxQUsEpZ2asqBx3fzYMuFRTFrMNKxn5B2QXAGMSuwVfEA9KSJr2CUs8vqbmhUCkzVvxysB4p6vybLS2CgQnze',
	},
	'prv': {
		'compressed': 'xprvA3KQP2Lsyt3wh7EQXbjaFefZ9whanHS8zBPbLRnXZaeWJ2Ni5AEiMzzPKbKHsG7Dn6hkhSkG8V4H4XUsjszxU4nd2sMnc5ag9sHLLYBqrr4',
		'segwit': 'yprvAMPaF7cjax34DFGCFZjPSPNwpLBQSfNzLubCPa2tHjEJeMLVSwe3i5uoEnUg4etxP3XEqr5ZJinjGkUJrte3xFNZ1jVKbjVaFJVHi4Msekw',
		'bech32': 'zprvAgruGXV5fS8bcAy51Yx2qCyDrT2kR6JjWcGMFHavoZiSaTXCpneXXNhvfsMLeyBmjzACqkpxB2KGCMAe85wUrW1dnenu6kVHCk4kXh7XFE6',
	}
}

# Source: Asgardex Wallet
vectors_multicoin = {
	'btc_bech32':   'bc1qwg77fxw0tkmc3h58tcnnpegxk7mp3h6ly44d3n',
	'eth':          '373731f4d885Fc7Da05498F9f0804a87A14F891b',
	'doge':         'DFX88RXpi4S4W24YVvuMgbdUcCAYNeEYGd',
	'avax-c':       '0x373731f4d885Fc7Da05498F9f0804a87A14F891b',
	'ltc_bech32':   'ltc1q3uh5ga5cp9kkdfx6a52uymxj9keq4tpzep7er0',
	'bch_compressed': 'bitcoincash:qpqpcllprftg4s0chdgkpxhxv23wfymq3gj7n0a9vw',
	'bsc_smart':    '0x373731f4d885Fc7Da05498F9f0804a87A14F891b',
	'bnb_beacon':   'bnb179c3ymltqm4utlp089zxqeta5dvn48a305rhe5',
	'rune':         'thor1nr6fye3nznyn20m5w6fey6w8a8l4q599cdqmpc',
}

def wif2addr(cfg, wif):
	from mmgen.tool.coin import tool_cmd
	return tool_cmd(
		cfg     = cfg.base_cfg,
		cmdname = 'wif2addr',
		proto   = cfg.base_cfg._proto,
		mmtype  = cfg.addr_type).wif2addr(wif)

class unit_tests:

	altcoin_deps = ('multicoin',)

	@property
	def _seed(self):
		if not hasattr(self, '__seed'):
			with open('test/ref/98831F3A.bip39') as fh:
				mnemonic = fh.read().strip()
			from mmgen.bip39 import bip39
			self.__seed = bip39().generate_seed(mnemonic.split())
		return self.__seed

	def chainparams(self, name, ut):
		for bipnum, idx, chain, addr_cls in (
				(44, 0,  'btc',  'P2PKH'),
				(49, 0,  'btc',  'P2SH'),
				(84, 0,  'btc',  'P2WPKH'),
				(44, 60, 'eth',  'Eth'),
				(44, 61, 'etc',  'Eth'),
				(44, 2,  'ltc',  'P2PKH'),
				(44, 3,  'doge', 'P2PKH'),
			):
			res = get_chain_params(bipnum, chain)
			assert res.idx == idx, res.idx
			assert res.chain == chain.upper()
			assert res.addr_cls == addr_cls
			vmsg(f'  {res}')
		vmsg('')
		return True

	def derive(self, name, ut):
		vmsg('seed: 98831F3A (default derivation)')

		m = MasterNode(cfg, self._seed)

		purpose = m.init_cfg(coin='btc', addr_type='bech32').derive_private()
		vmsg(f'  {purpose.address=}')

		coin_type1 = purpose.derive_private()

		coin_type2 = m.to_coin_type(coin='btc', addr_type='bech32')
		assert coin_type1.address == coin_type2.address
		vmsg(f'  {coin_type1.address=}')

		acct = coin_type2.derive_private(idx=0)
		chain1 = acct.derive_private(idx=0, hardened=False)

		chain2 = m.to_chain(idx=0, coin='btc', addr_type='bech32', public=False)
		assert chain2.address == chain1.address

		chain3 = m.to_coin_type(coin='btc', addr_type='bech32').to_chain(0, public=True)
		assert chain3.address == chain1.address
		vmsg(f'  {chain1.address=}')

		a = BipHDNode.from_extended_key(cfg, 'btc', chain2.xpub)
		b = BipHDNode.from_extended_key(cfg, 'btc', chain2.xprv)
		vmsg(
			'\n  xpub:\n' +
			fmt(str(Bip32ExtendedKey(b.xpub)), indent='    ')
		)
		assert a.xpub == b.xpub

		vmsg('  Addresses:')
		for i in range(3):
			res = chain1.derive_public(i)
			vmsg(f'    {i} {res.address}')
			assert res.address == vectors_derive['bech32'][i]
			res = chain1.derive_private(i)
			assert res.address == vectors_derive['bech32'][i]

		vmsg('')
		return True

	def derive_addrfmt(self, name, ut):
		vmsg('seed: 98831F3A (default derivation)')

		m = MasterNode(cfg, self._seed)

		for addr_type in ('compressed', 'segwit', 'bech32'):
			chk_xpub = vectors_addrfmt['pub'][addr_type]
			chk_xprv = vectors_addrfmt['prv'][addr_type]

			res1 = m.to_chain(idx=0, coin='btc', addr_type=addr_type).derive_public(0)
			vmsg(f'  {addr_type}: {res1.xpub}')
			assert res1.xpub == chk_xpub

			res2 = m.to_chain(idx=0, coin='btc', addr_type=addr_type).derive_private(0, False)
			vmsg(f'  {addr_type}: {res2.xprv}')
			assert res2.xprv == chk_xprv
			assert res2.xpub == chk_xpub

			assert res2.address == wif2addr(res2.cfg, res2.privkey.wif)

		vmsg('')
		return True

	def path(self, name, ut):

		for vec in vectors_bip32:
			seed = bytes.fromhex(vec['seed'])
			vmsg(f'Seed: {vec["seed"]}')

			for n, path_str in enumerate(vec):
				if path_str in ('seed', 'comment'):
					continue

				path_arg = path_str.replace("'", 'H') if n % 2 else path_str
				node = BipHDNode.from_path(cfg, seed, path_arg, no_path_checks=True)
				vmsg('  Path {} {}'.format(pink(path_str), blue('('+node.desc+')')))

				for xkey_type in ('xpub', 'xprv'):
					vmsg(f'    {getattr(node, xkey_type)}')
					assert getattr(node, xkey_type) == vec[path_str][xkey_type]

			vmsg('')

		return True

	def parse_extended(self, name, ut):
		vmsg('Parsing and validating extended keys:\n')

		for vec in vectors_bip32:
			vmsg(f'  Seed: {vec["seed"]}')

			for path_str in vec:
				if path_str in ('seed', 'comment'):
					continue

				vmsg('    Path {}'.format(pink(path_str)))
				for xkey_type in ('xpub', 'xprv'):
					xkey = vec[path_str][xkey_type]
					vmsg(f'      {xkey}')
					node = BipHDNode.from_extended_key(cfg, 'btc', xkey)
					assert getattr(node, xkey_type) == xkey

			vmsg('')

		return True

	def multicoin(self, name, ut):
		m = MasterNode(cfg, self._seed)

		fs = '  {:6} {:10} {}'
		vmsg(fs.format('COIN', 'ADDR_TYPE', 'ADDR'))
		for id_str, addr_chk in vectors_multicoin.items():
			ss = id_str.split('_')
			coin = ss[0]
			addr_type = ss[1] if len(ss) == 2 else None
			if coin not in BipHDConfig.supported_coins:
				vmsg(gray(fs.format(coin.upper(), (addr_type or ''), '[not supported yet]')))
				continue
			vmsg(fs.format(coin.upper(), (addr_type or 'auto'), addr_chk))
			node = m.to_chain(idx=0, coin=coin, addr_type=addr_type).derive_private(0)
			xpub_parsed = node.key_extended(public=True)
			xprv_parsed = node.key_extended(public=False)
			addr = node.address
			at_arg = 'compressed' if coin == 'doge' else None
			from_xpub = BipHDNode.from_extended_key(node.cfg.base_cfg, coin, xpub_parsed.base58, addr_type=at_arg)
			from_xprv = BipHDNode.from_extended_key(node.cfg.base_cfg, coin, xprv_parsed.base58, addr_type=at_arg)
			assert from_xpub.xpub == node.xpub, f'{from_xpub.xpub=} != {node.xpub}'
			assert from_xprv.xpub == node.xpub, f'{from_xprv.xpub=} != {node.xpub}'
			assert from_xpub.address == addr, f'{from_xpub.address} != {addr}'
			assert from_xprv.address == addr, f'{from_xprv.address} != {addr}'
			addr_from_wif = wif2addr(node.cfg, node.privkey.wif)
			proto = node.cfg.base_cfg._proto
			if proto.base_proto == 'Ethereum':
				addr = proto.checksummed_addr(node.address)
				addr_from_wif = proto.checksummed_addr(addr_from_wif)
			assert addr == addr_chk, f'{addr} != {addr_chk}'
			assert addr == addr_from_wif, f'{addr} != {addr_from_wif}'

		vmsg('')
		return True

	def errors(self, name, ut):
		vmsg('Checking error handling:')

		m = MasterNode(cfg, self._seed)
		m_btc = m.init_cfg(coin='btc', addr_type='bech32')

		purpose = m_btc.derive_private()
		coin_type = purpose.derive_private()
		acct = coin_type.derive_private(idx=0)
		chain = acct.derive_private(idx=0, hardened=False)

		def bad01():
			m.to_chain(idx=0, coin='erq', addr_type='C')
		def bad02():
			m_btc.derive_private(idx=0)
		def bad03():
			m_btc.derive_private(hardened=False)
		def bad04():
			purpose.derive_private(idx=8)
		def bad05():
			purpose.derive_private(hardened=False)
		def bad06():
			coin_type.derive_private() # no acct idx
		def bad08():
			m_btc.derive_public() # must be private
		def bad09():
			coin_type.derive_private(idx=8, hardened=False)
		def bad10():
			acct.derive_private()
		def bad11():
			chain.derive_private()
		def bad12():
			chain.derive_private(hardened=True, idx=3)

		bad_data = (
			('unsupported coin',                       'ValueError', 'not supported',        bad01),
			('depth 1 (purpose):   idx not None',      'ValueError', 'index for path comp',  bad02),
			('depth 1 (purpose):   hardened False',    'ValueError', 'value for ‘hardened’', bad03),
			('depth 2 (coin type): idx mismatch',      'ValueError', 'index 8 at depth',     bad04),
			('depth 2 (coin type): hardened False',    'ValueError', 'value for ‘hardened’', bad05),
			('depth 3 (account):   idx not set',       'ValueError', 'must be set',          bad06),
			('depth 1 (purpose):   node not hardened', 'ValueError', 'must be hardened',     bad08),
			('depth 3 (account):   node not hardened', 'ValueError', 'value for ‘hardened’', bad09),
			('depth 4 (chain):     idx not set',       'ValueError', 'must be either 0',     bad10),
			('depth 5 (leaf node): idx not set',       'ValueError', 'must be set',          bad11),
			('depth 5 (leaf node): hardened True',     'ValueError', 'must be None',         bad12),
		)

		ut.process_bad_data(bad_data, pfx='')
		vmsg('')
		return True

	def parse_extended_errors(self, name, ut):
		vmsg('Parsing and validating invalid extended keys:')
		vec = vectors_bip32_invalid
		func = [lambda m=n: BipHDNode.from_extended_key(cfg, 'btc', vec[m][0]) for n in range(len(vec))]
		exc = (
			'first byte for public',
			'first byte for private',
			'first byte for public',
			'first byte for private',
			'first byte for public',
			'first byte for private',
			'non-zero parent fingerprint',
			'non-zero parent fingerprint',
			'non-zero index',
			'non-zero index',
			'unrecognized extended key v',
			'unrecognized extended key v',
			'private key is zero!',
			'private key >= group order!',
			'Public key could not be parsed', # extmod
			'incorrect checksum',
		)
		ut.process_bad_data([(vec[n][1], 'ValueError', exc[n], func[n]) for n in range(len(vec))], pfx='')
		vmsg('')
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet


"""
test.cmdtest_d.xmr_autosign: xmr autosigning tests for the cmdtest.py test suite
"""

import re

from mmgen.color import blue, cyan, brown
from mmgen.util import async_run

from ..include.common import imsg, silence, end_silence
from .include.common import get_file_with_ext

from .xmrwallet import CmdTestXMRWallet
from .autosign import CmdTestAutosignThreaded

def make_burn_addr(cfg):
	from mmgen.tool.coin import tool_cmd
	return tool_cmd(
		cfg     = cfg,
		cmdname = 'privhex2addr',
		proto   = cfg._proto,
		mmtype  = 'monero').privhex2addr('beadcafe'*8)

class CmdTestXMRAutosign(CmdTestXMRWallet, CmdTestAutosignThreaded):
	"""
	Monero autosigning operations
	"""
	tmpdir_nums = [39]

	# xmrwallet attrs:
	tx_relay_user = 'miner'
	#    user     sid      autosign  shift kal_range add_coind_args
	user_data = (
		('miner', '98831F3A', False, 130, '1', []),
		('alice', 'FE3C6545', True,  150, '1-2', []),
	)

	# autosign attrs:
	coins = ['xmr']

	cmd_group = (
		('daemon_version',           'checking daemon version'),
		('gen_kafile_miner',         'generating key-address file for Miner'),
		('create_wallet_miner',      'creating Monero wallet for Miner'),
		('mine_initial_coins',       'mining initial coins'),
		('create_tmp_wallets',       'creating temporary online wallets for Alice'),
		('new_account_alice',        'adding an account to Alice’s tmp wallet'),
		('new_address_alice',        'adding an address to Alice’s tmp wallet'),
		('new_address_alice_label',  'adding an address to Alice’s tmp wallet (with label)'),
		('dump_tmp_wallets',         'dumping Alice’s tmp wallets'),
		('delete_tmp_wallets',       'deleting Alice’s tmp wallets'),
		('autosign_setup',           'autosign setup with Alice’s seed'),
		('autosign_xmr_setup',       'autosign setup (creation of Monero signing wallets)'),
		('create_watchonly_wallets', 'creating watch-only wallets from Alice’s wallet dumps'),
		('delete_tmp_dump_files',    'deleting Alice’s dump files'),
		('fund_alice1',              'sending funds to Alice (wallet #1)'),
		('check_bal_alice1',         'mining, checking balance (wallet #1)'),
		('fund_alice2',              'sending funds to Alice (wallet #2)'),
		('check_bal_alice2',         'mining, checking balance (wallet #2)'),
		('wait_loop_start',          'starting autosign wait loop'),
		('export_outputs1',          'exporting outputs from Alice’s watch-only wallet #1'),
		('create_transfer_tx1',      'creating a transfer TX'),
		('submit_transfer_tx1',      'submitting the transfer TX'),
		('resubmit_transfer_tx1',    'resubmitting the transfer TX'),
		('export_outputs2',          'exporting outputs from Alice’s watch-only wallet #1'),
		('import_key_images1',       'importing signed key images into Alice’s online wallets'),
		('sync_chkbal1',             'syncing Alice’s wallet #1'),
		('abort_tx1',                'aborting the current transaction (error)'),
		('create_transfer_tx2',      'creating a transfer TX (for relaying via proxy)'),
		('abort_tx2',                'aborting the current transaction (OK, unsigned)'),
		('create_transfer_tx2a',     'creating the transfer TX again'),
		('submit_transfer_tx2',      'submitting the transfer TX (relaying via proxy)'),
		('sync_chkbal2',             'syncing Alice’s wallets and checking balance'),
		('dump_wallets',             'dumping Alice’s wallets'),
		('delete_wallets',           'deleting Alice’s wallets'),
		('restore_wallets',          'creating online (watch-only) wallets for Alice'),
		('delete_dump_files',        'deleting Alice’s dump files'),
		('export_outputs3',          'exporting outputs from Alice’s watch-only wallets'),
		('import_key_images2',       'importing signed key images into Alice’s online wallets'),
		('sync_chkbal3',             'syncing Alice’s wallets and checking balance'),
		('wait_loop_kill',           'stopping autosign wait loop'),
		('stop_daemons',             'stopping all wallet and coin daemons'),
		('view',                     'viewing Alice’s wallet in offline mode (wallet #1)'),
		('listview',                 'list-viewing Alice’s wallet in offline mode (wallet #2)'),
		('txlist',                   'listing Alice’s submitted transactions'),
		('txview',                   'viewing Alice’s submitted transactions'),
		('txview_all',               'viewing all raw, signed and submitted transactions'),
		('check_tx_dirs',            'cleaning and checking signable file directories'),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):

		CmdTestAutosignThreaded.__init__(self, cfg, trunner, cfgs, spawn)
		CmdTestXMRWallet.__init__(self, cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		from mmgen.cfg import Config
		self.alice_cfg = Config({
			'coin': 'XMR',
			'outdir': self.users['alice'].udir,
			'wallet_dir': self.users['alice'].udir,
			'wallet_rpc_password': 'passwOrd',
			'test_suite': True,
		})

		self.burn_addr = make_burn_addr(cfg)

		self.opts.append('--xmrwallets={}'.format(self.users['alice'].kal_range)) # mmgen-autosign opts
		self.autosign_opts = ['--autosign']                                       # mmgen-xmrwallet opts
		self.spawn_env['MMGEN_TEST_SUITE_XMR_AUTOSIGN'] = '1'

	def create_tmp_wallets(self):
		self.spawn(msg_only=True)
		data = self.users['alice']
		from mmgen.wallet import Wallet
		from mmgen.xmrwallet import op
		from mmgen.addrlist import KeyAddrList
		silence()
		kal = KeyAddrList(
			cfg       = self.alice_cfg,
			proto     = self.proto,
			addr_idxs = '1-2',
			seed      = Wallet(self.alice_cfg, fn=data.mmwords).seed,
			skip_chksum_msg = True,
			key_address_validity_check = False)
		kal.file.write(ask_overwrite=False)
		fn = get_file_with_ext(data.udir, 'akeys')
		m = op('create', self.alice_cfg, fn, '1-2')
		async_run(m.main())
		async_run(m.stop_wallet_daemon())
		end_silence()
		return 'ok'

	def _new_addr_alice(self, *args):
		data = self.users['alice']
		return self.new_addr_alice(
			*args,
			kafile = get_file_with_ext(data.udir, 'akeys'))

	def new_account_alice(self):
		return self._new_addr_alice(
			'2',
			'start',
			r'Creating new account for wallet .*2.* with label .*‘xmrwallet new account .*y/N\): ')

	def new_address_alice(self):
		return self._new_addr_alice(
			'2:1',
			'continue',
			r'Creating new address for wallet .*2.*, account .*#1.* with label .*‘xmrwallet new address .*y/N\): ')

	def new_address_alice_label(self):
		return self._new_addr_alice(
			'2:1,Alice’s new address',
			'stop',
			r'Creating new address for wallet .*2.*, account .*#1.* with label .*‘Alice’s new address .*y/N\): ')

	def dump_tmp_wallets(self):
		return self._dump_wallets(autosign=False)

	def dump_wallets(self):
		return self._dump_wallets(autosign=True)

	def _dump_wallets(self, autosign):
		data = self.users['alice']
		self.insert_device_online()
		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ [f'--wallet-dir={data.udir}', f'--daemon=localhost:{data.md.rpc_port}']
			+ (self.autosign_opts if autosign else [])
			+ ['dump']
			+ ([] if autosign else [get_file_with_ext(data.udir, 'akeys')]))
		t.expect('2 wallets dumped')
		t.read()
		self.remove_device_online()
		return t

	def _delete_files(self, *ext_list):
		data = self.users['alice']
		self.spawn(msg_only=True)
		for ext in ext_list:
			get_file_with_ext(data.udir, ext, no_dot=True, delete_all=True)
		return 'ok'

	def delete_tmp_wallets(self):
		return self._delete_files('MoneroWallet', 'MoneroWallet.keys', '.akeys')

	def delete_wallets(self):
		return self._delete_files('MoneroWatchOnlyWallet', '.keys', '.address.txt')

	def delete_tmp_dump_files(self):
		return self._delete_files('.dump')

	def gen_kafile_miner(self):
		return self.gen_kafiles(['miner'])

	def create_wallet_miner(self):
		return self.create_wallets_miner()

	def delete_dump_files(self):
		return self._delete_files('.dump')

	def fund_alice1(self):
		return self.fund_alice(wallet=1)

	def check_bal_alice1(self):
		return self.check_bal_alice(wallet=1)

	def fund_alice2(self):
		return self.fund_alice(wallet=2)

	def check_bal_alice2(self):
		return self.check_bal_alice(wallet=2)

	def autosign_setup(self):
		return self.run_setup(
			mn_type        = 'mmgen',
			mn_file        = self.users['alice'].mmwords,
			use_dfl_wallet = None,
			expect_args    = ['Continue with Monero setup? (Y/n): ', 'n'])

	def autosign_xmr_setup(self):
		self.insert_device_online()
		self.do_mount_online()
		self.asi_online.xmr_dir.mkdir(exist_ok=True)
		(self.asi_online.xmr_dir / 'old.vkeys').touch()
		self.do_umount_online()
		self.remove_device_online()

		self.insert_device()
		t = self.spawn('mmgen-autosign', self.opts + ['xmr_setup'], no_passthru_opts=True)
		t.written_to_file('View keys')
		t.read()
		self.remove_device()
		return t

	def create_watchonly_wallets(self):
		return self.restore_wallets()

	def restore_wallets(self):
		self.insert_device_online()
		t = self.create_wallets('alice', op='restore')
		t.read() # required!
		self.remove_device_online()
		return t

	def _create_transfer_tx(self, amt, add_opts=[]):
		self.insert_device_online()
		t = self.do_op(
			'transfer',
			'alice',
			f'1:0:{self.burn_addr},{amt}',
			no_relay = True,
			do_ret   = True,
			add_opts = add_opts)
		t.read() # required!
		self.remove_device_online()
		return t

	def create_transfer_tx1(self):
		return self._create_transfer_tx('0.124', add_opts=['--priority=2'])

	def create_transfer_tx2(self):
		return self._create_transfer_tx('0.257')

	create_transfer_tx2a = create_transfer_tx2

	def _abort_tx(self, expect, send=None, exit_val=None):
		self.insert_device_online()
		t = self.spawn('mmgen-xmrwallet', ['--autosign', 'abort'], exit_val=exit_val)
		t.expect(expect)
		if send:
			t.send(send)
		t.read() # required!
		self.remove_device_online()
		return t

	def abort_tx1(self):
		return self._abort_tx('No unsent transactions present', exit_val=2)

	def abort_tx2(self):
		return self._abort_tx('(y/N): ', 'y')

	def _xmr_autosign_op(
			self,
			op,
			desc          = None,
			signable_desc = None,
			ext           = None,
			wallet_arg    = None,
			add_opts      = [],
			wait_signed   = False):
		if wait_signed:
			self._wait_signed(signable_desc)
		data = self.users['alice']
		args = (
			self.extra_opts
			+ self.autosign_opts
			+ [f'--wallet-dir={data.udir}', f'--daemon=localhost:{data.md.rpc_port}']
			+ add_opts
			+ [op]
			+ ([get_file_with_ext(self.asi.xmr_tx_dir, ext)] if ext else [])
			+ ([wallet_arg] if wallet_arg else [])
		)
		desc_pfx = f'{desc}, ' if desc else ''
		self.insert_device_online() # device must be removed by calling method
		return self.spawn('mmgen-xmrwallet', args, extra_desc=f'({desc_pfx}Alice)')

	def _sync_chkbal(self, wallet_arg, bal_chk_func):
		return self.sync_wallets(
			'alice',
			op           = 'sync',
			wallets      = wallet_arg,
			bal_chk_func = bal_chk_func)

	def sync_chkbal1(self):
		return self._sync_chkbal('1', lambda n, b, ub: b == ub and 1 < b < 1.12)
		# 1.234567891234 - 0.124 = 1.110567891234 (minus fees)

	def sync_chkbal2(self):
		return self._sync_chkbal('1', lambda n, b, ub: b == ub and 0.8 < b < 0.86)
		# 1.234567891234 - 0.124 - 0.257 = 0.853567891234 (minus fees)

	def sync_chkbal3(self):
		return self._sync_chkbal(
			'1-2',
			lambda n, b, ub: b == ub and ((n == 1 and 0.8 < b < 0.86) or (n == 2 and b > 1.23)))

	def _mine_chk(self, desc):
		bal_type = {'locked':'b', 'unlocked':'ub'}[desc]
		return self.mine_chk(
			'alice', 1, 0,
			lambda x: 0 < getattr(x, bal_type) < 1.234567891234,
			f'{desc} balance 0 < 1.234567891234')

	def submit_transfer_tx1(self):
		return self._submit_transfer_tx()

	def resubmit_transfer_tx1(self):
		return self._submit_transfer_tx(
				relay_parm = self.tx_relay_daemon_proxy_parm,
				op         = 'resubmit',
				check_bal  = False)

	def submit_transfer_tx2(self):
		return self._submit_transfer_tx(relay_parm=self.tx_relay_daemon_parm)

	def _submit_transfer_tx(self, relay_parm=None, ext=None, op='submit', check_bal=True):
		t = self._xmr_autosign_op(
			op            = op,
			add_opts      = [f'--tx-relay-daemon={relay_parm}'] if relay_parm else [],
			ext           = ext,
			signable_desc = 'transaction',
			wait_signed   = op == 'submit')
		t.expect(f'{op.capitalize()} transaction? (y/N): ', 'y')
		t.written_to_file('Submitted transaction')
		t.read()
		self.remove_device_online() # device was inserted by _xmr_autosign_op()
		if check_bal:
			t.ok()
			return self._mine_chk('unlocked')
		else:
			return t

	def _export_outputs(self, wallet_arg, op, add_opts=[]):
		t = self._xmr_autosign_op(
			op         = op,
			wallet_arg = wallet_arg,
			add_opts   = add_opts)
		t.written_to_file('Wallet outputs')
		t.read()
		self.remove_device_online() # device was inserted by _xmr_autosign_op()
		return t

	def export_outputs1(self):
		return self._export_outputs('1', op='export-outputs')

	def export_outputs2(self): # NB: --rescan-spent does not work with testnet/stagenet
		return self._export_outputs('1', op='export-outputs-sign', add_opts=['--rescan-blockchain'])

	def export_outputs3(self):
		return self._export_outputs('1-2', op='export-outputs-sign')

	def _import_key_images(self, wallet_arg):
		t = self._xmr_autosign_op(
			op            = 'import-key-images',
			wallet_arg    = wallet_arg,
			signable_desc = 'wallet outputs',
			wait_signed   = True)
		t.read()
		self.remove_device_online() # device was inserted by _xmr_autosign_op()
		return t

	def import_key_images1(self):
		return self._import_key_images(None)

	def import_key_images2(self):
		return self._import_key_images(None)

	def txlist(self):
		self.insert_device_online()
		t = self.spawn('mmgen-xmrwallet', self.autosign_opts + ['txlist'])
		t.match_expect_list([
			'SUBMITTED',
			'Network', 'Submitted',
			'transfer 1:0', '-> ext',
			'transfer 1:0', '-> ext'
		])
		t.read()
		self.remove_device_online()
		return t

	def txview(self):
		self.insert_device_online()
		t = self.spawn('mmgen-xmrwallet', self.autosign_opts + ['txview'])
		t.read()
		self.remove_device_online()
		return t

	def txview_all(self):
		self.spawn(msg_only=True)
		self.insert_device()
		self.do_mount()
		imsg(blue('Opening transaction directory: ') + cyan(f'{self.asi.xmr_tx_dir}'))
		for fn in self.asi.xmr_tx_dir.iterdir():
			imsg('\n' + brown(f'Viewing ‘{fn.name}’'))
			self.spawn('mmgen-xmrwallet', ['txview', str(fn)], no_msg=True).read()
		imsg('')
		self.do_umount()
		self.remove_device()
		return 'ok'

	def check_tx_dirs(self):

		self.insert_device()
		self.do_mount()
		before = '\n'.join(self._gen_listing())
		self.do_umount()
		self.remove_device()

		self.insert_device()
		t = self.spawn('mmgen-autosign', self.opts + ['clean'])
		t.read()
		self.remove_device()

		self.insert_device()
		self.do_mount()
		after = '\n'.join(self._gen_listing())
		self.do_umount()
		self.remove_device()

		imsg(f'\nBefore cleaning:\n{before}')
		imsg(f'\nAfter cleaning:\n{after}')
		pat = r'xmr/tx: \s*\S+\.subtx \S+\.subtx\s+xmr/outputs:\s*$'
		assert re.search(pat, after, re.DOTALL), f'regex search for {pat} failed'
		return t

	def view(self):
		return self.sync_wallets('alice', op='view', wallets='1')

	def listview(self):
		return self.sync_wallets('alice', op='listview', wallets='2')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.signed: THORChain signed transaction class
"""

from ....tx import signed as TxBase
from ....obj import CoinTxID, NonNegativeInt
from .completed import Completed

class Signed(Completed, TxBase.Signed):

	desc = 'signed transaction'

	def parse_txfile_serialized_data(self):
		from .protobuf import RuneTx
		tx = RuneTx.loads(bytes.fromhex(self.serialized))

		b = tx.body.messages[0].body
		i = tx.authInfo

		from_k, amt_k = ('signer', 'coins') if self.is_swap else ('fromAddress', 'amount')

		self.txobj = {
			'from':     self.proto.encode_addr_bech32x(getattr(b, from_k)),
			'to':       None if self.is_swap else self.proto.encode_addr_bech32x(b.toAddress),
			'amt':      self.proto.coin_amt(int(getattr(b, amt_k)[0].amount), from_unit='satoshi'),
			'gas':      NonNegativeInt(i.fee.gasLimit),
			'sequence': NonNegativeInt(i.signerInfos[0].sequence)}

		txid = CoinTxID(tx.txid)
		assert txid == self.coin_txid, 'serialized txid doesn’t match txid in MMGen transaction file'

class AutomountSigned(TxBase.AutomountSigned, Signed):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addrfile: Address and password file classes for the MMGen suite
"""

from .cfg import gc
from .util import msg, die, capfirst
from .protocol import init_proto
from .obj import MMGenObject, TwComment, WalletPassword, MMGenPWIDString
from .seed import SeedID, is_seed_id
from .key import PrivKey
from .addr import ViewKey, AddrListID, MMGenAddrType, MMGenPasswordType, is_addr_idx
from .addrlist import KeyList, AddrListData

class AddrFile(MMGenObject):
	desc        = 'addresses'
	ext         = 'addrs'
	line_ctr    = 0
	header = """
# This file is editable.
# Everything following a hash symbol '#' is a comment and ignored by {pnm}.
"""
	text_label_header = """
# A text label of {n} screen cells or less may be added to the right of each
# address, and it will be appended to the tracking wallet label upon import.
# The label may contain any printable ASCII symbol.
"""

	def __init__(self, parent):
		self.parent = parent
		self.cfg    = parent.cfg
		self.infile = None
		self.fmt_data = None

	def encrypt(self):
		from .crypto import Crypto
		self.fmt_data = Crypto(self.cfg).mmgen_encrypt(
			data = self.fmt_data.encode(),
			desc = f'new {self.parent.desc} list')
		self.ext += f'.{Crypto.mmenc_ext}'

	@property
	def filename(self):
		return '{}{}.{}'.format(
			self.parent.id_str,
			('.' + self.parent.proto.network) if self.parent.proto.testnet else '',
			self.ext)

	def write(
			self,
			fn            = None,
			*,
			binary        = False,
			desc          = None,
			ask_overwrite = True,
			outdir        = None):
		from .fileutil import write_data_to_file
		write_data_to_file(
			cfg           = self.cfg,
			outfile       = fn or self.filename,
			data          = self.fmt_data or self.format(),
			desc          = desc or self.desc,
			ask_tty       = self.parent.has_keys and not self.cfg.quiet,
			binary        = binary,
			ask_overwrite = ask_overwrite,
			outdir        = outdir)

	def make_label(self):
		proto = self.parent.proto
		coin = proto.coin
		mmtype = self.parent.al_id.mmtype
		lbl_p2 = ':'.join(
			([] if coin == 'BTC' or (coin == 'BCH' and not self.cfg.cashaddr) else [coin])
			+ ([] if mmtype == 'E' or (mmtype == 'L' and not proto.testnet) else [mmtype.name.upper()])
			+ ([proto.network.upper()] if proto.testnet else [])
		)
		return self.parent.al_id.sid + (' ' if lbl_p2 else '') + lbl_p2

	def format(self, *, add_comments=False):
		p = self.parent
		if p.gen_passwds and p.pw_fmt in ('bip39', 'xmrseed'):
			desc_pfx = f'{p.pw_fmt.upper()} '
			hdr2 = ''
		else:
			desc_pfx = ''
			hdr2 = self.text_label_header
		out = [
			f'# {gc.proj_name} {desc_pfx}{p.desc} file\n#\n'
			+ self.header.strip().format(pnm=gc.proj_name)
			+ '\n'
			+ hdr2.lstrip().format(n=TwComment.max_screen_width)
			+ '#\n'
		]

		if p.chksum:
			out.append(f'# {capfirst(p.desc)} data checksum for {p.id_str}: {p.chksum}')
			out.append('# Record this value to a secure location.\n')

		lbl = self.make_label()
		self.parent.dmsg_sc('lbl', lbl[9:])
		out.append(f'{lbl} {{')

		fs = '  {:<%s}  {:<34}{}' % len(str(p.data[-1].idx))
		for e in p.data:
			c = ' ' + e.comment if add_comments and e.comment else ''
			if type(p) is KeyList:
				out.append(fs.format(e.idx, f'{p.al_id.mmtype.wif_label}: {e.sec.wif}', c))
			elif type(p).__name__ == 'PasswordList':
				out.append(fs.format(e.idx, e.passwd, c))
			else: # First line with idx
				out.append(fs.format(e.idx, e.addr.views[e.addr.view_pref], c))
				if p.has_keys:
					if self.cfg.b16:
						out.append(fs.format('', f'orig_hex: {e.sec.orig_bytes.hex()}', c))
					if type(self) is not ViewKeyAddrFile:
						out.append(fs.format('', f'{p.al_id.mmtype.wif_label}: {e.sec.wif}', c))
					for k in ('viewkey', 'wallet_passwd'):
						v = getattr(e, k)
						if v:
							out.append(fs.format('', f'{k}: {v}', c))

		out.append('}')
		self.fmt_data = '\n'.join([l.rstrip() for l in out]) + '\n'
		return self.fmt_data

	def get_line(self, lines):
		ret = lines.pop(0).split(None, 2)
		self.line_ctr += 1
		if ret[0] == 'orig_hex:': # hacky
			ret = lines.pop(0).split(None, 2)
			self.line_ctr += 1
		return ret if len(ret) == 3 else ret + ['']

	def parse_file_body(self, lines):

		p = self.parent
		ret = AddrListData()
		le = p.entry_type
		iifs = "{!r}: invalid identifier [expected '{}:']"

		while lines:
			idx, addr, comment = self.get_line(lines)

			assert is_addr_idx(idx), f'invalid address index {idx!r}'
			p.check_format(addr)

			a = le(**{'proto': p.proto, 'idx':int(idx), p.main_attr:addr, 'comment':comment})

			if p.has_keys: # order: wif, (orig_hex), viewkey, wallet_passwd
				if type(self) is not ViewKeyAddrFile:
					d = self.get_line(lines)
					assert d[0] == p.al_id.mmtype.wif_label+':', iifs.format(d[0], p.al_id.mmtype.wif_label)
					a.sec = PrivKey(proto=p.proto, wif=d[1])
				for k, dtype, add_proto in (
					('viewkey', ViewKey, True),
					('wallet_passwd', WalletPassword, False)):
					if k in p.al_id.mmtype.extra_attrs:
						d = self.get_line(lines)
						assert d[0] == k+':', iifs.format(d[0], k)
						setattr(a, k, dtype(*((p.proto, d[1]) if add_proto else (d[1],))))

			ret.append(a)

		if type(self) is not ViewKeyAddrFile and p.has_keys and p.ka_validity_chk is not False:

			def verify_keys():
				from .addrgen import KeyGenerator, AddrGenerator
				kg = KeyGenerator(self.cfg, p.proto, p.al_id.mmtype.pubkey_type)
				ag = AddrGenerator(self.cfg, p.proto, p.al_id.mmtype)
				llen = len(ret)
				qmsg_r = p.cfg._util.qmsg_r
				for n, e in enumerate(ret):
					qmsg_r(f'\rVerifying keys {n+1}/{llen}')
					assert e.addr == ag.to_addr(kg.gen_data(e.sec)), (
						f'Key doesn’t match address!\n  {e.sec.wif}\n  {e.addr}')
				p.cfg._util.qmsg(' - done')

			if self.cfg.yes or p.ka_validity_chk:
				verify_keys()
			else:
				from .ui import keypress_confirm
				if keypress_confirm(p.cfg, 'Check key-to-address validity?'):
					verify_keys()

		return ret

	def parse_file(self, fn, *, buf=[], exit_on_error=True):

		def parse_addrfile_label(lbl):
			"""
			label examples:
			- Bitcoin legacy mainnet:           no label
			- BCH legacy mainnet (no cashaddr): no label
			- BCH legacy mainnet (cashaddr):    'BCH'
			- Bitcoin legacy testnet:           'LEGACY:TESTNET'
			- Bitcoin Segwit:                   'SEGWIT'
			- Bitcoin Segwit testnet:           'SEGWIT:TESTNET'
			- Bitcoin Bech32 regtest:           'BECH32:REGTEST'
			- Litecoin legacy mainnet:          'LTC'
			- Litecoin Bech32 mainnet:          'LTC:BECH32'
			- Litecoin legacy testnet:          'LTC:LEGACY:TESTNET'
			- Ethereum mainnet:                 'ETH'
			- Ethereum Classic mainnet:         'ETC'
			- Ethereum regtest:                 'ETH:REGTEST'
			"""
			lbl = lbl.lower()

			# remove the network component:
			if lbl.endswith(':testnet'):
				network = 'testnet'
				lbl = lbl[:-8]
			elif lbl.endswith(':regtest'):
				network = 'regtest'
				lbl = lbl[:-8]
			else:
				network = 'mainnet'

			from .proto.btc.params import mainnet
			if lbl in [MMGenAddrType(mainnet, key).name for key in mainnet.mmtypes]:
				coin, mmtype_key = ('BTC', lbl)
			elif ':' in lbl: # first component is coin, second is mmtype_key
				coin, mmtype_key = lbl.split(':')
			else:            # only component is coin
				coin, mmtype_key = (lbl, None)

			proto = init_proto(p.cfg, coin=coin, network=network)

			if mmtype_key is None:
				mmtype_key = proto.mmtypes[0]

			return (proto, proto.addr_type(mmtype_key))

		p = self.parent

		from .fileutil import get_lines_from_file
		lines = get_lines_from_file(p.cfg, fn, desc=f'{p.desc} data', trim_comments=True)

		try:
			assert len(lines) >= 3, f'Too few lines in address file ({len(lines)})'
			ls = lines[0].split()
			assert 1 < len(ls) < 5, f'Invalid first line for {p.gen_desc} file: {lines[0]!r}'
			assert ls[-1] == '{', f'{ls!r}: invalid first line'
			ls.pop()
			assert lines[-1] == '}', f'{lines[-1]!r}: invalid last line'
			sid = ls.pop(0)
			assert is_seed_id(sid), f'{sid!r}: invalid Seed ID'

			if type(p).__name__ == 'PasswordList' and len(ls) == 2:
				ss = ls.pop().split(':')
				assert len(ss) == 2, f'{ss!r}: invalid password length specifier (must contain colon)'
				p.set_pw_fmt(ss[0])
				p.set_pw_len(ss[1])
				p.pw_id_str = MMGenPWIDString(ls.pop())
				modname, funcname = p.pw_info[p.pw_fmt].chk_func.split('.')
				import importlib
				p.chk_func = getattr(importlib.import_module('mmgen.'+modname), funcname)
				proto = init_proto(p.cfg, 'btc') # FIXME: dummy protocol
				mmtype = MMGenPasswordType(proto, 'P')
			elif len(ls) == 1:
				proto, mmtype = parse_addrfile_label(ls[0])
			elif len(ls) == 0:
				proto = init_proto(p.cfg, 'btc')
				mmtype = proto.addr_type('L')
			else:
				raise ValueError(f'{lines[0]}: Invalid first line for {p.gen_desc} file {fn!r}')

			if type(p).__name__ != 'PasswordList':
				if proto.base_coin != p.proto.base_coin or proto.network != p.proto.network:
					# Having caller supply protocol and checking address file protocol against it here
					# allows us to catch all mismatches in one place.  This behavior differs from that of
					# transaction files, which determine the protocol independently, requiring the caller
					# to check for protocol mismatches (e.g. mmgen.tx.completed.check_correct_chain())
					raise ValueError(
						f'{p.desc} file is '
						+ f'{proto.base_coin} {proto.network} but protocol is '
						+ f'{p.proto.base_coin} {p.proto.network}')

			p.base_coin = proto.base_coin
			p.network = proto.network
			p.al_id = AddrListID(sid=SeedID(sid=sid), mmtype=mmtype)

			data = self.parse_file_body(lines[1:-1])
			assert isinstance(data, list), 'Invalid file body data'
		except Exception as e:
			m_add = f', content line {self.line_ctr}' if self.line_ctr else ''
			m = f'Invalid data in {p.desc} list file ‘{fn}’{m_add} ({e!s})'
			if exit_on_error:
				die(3, m)
			else:
				msg(m)
				return False

		return data

class KeyAddrFile(AddrFile):
	desc = 'secret keys'
	ext  = 'akeys'

class ViewKeyAddrFile(KeyAddrFile):
	desc = 'view keys'
	ext  = 'vkeys'

class KeyFile(KeyAddrFile):
	ext         = 'keys'
	header = """
# This file is editable.
# Everything following a hash symbol '#' is a comment and ignored by {pnm}.
"""
	text_label_header = ''

class PasswordFile(AddrFile):
	desc        = 'passwords'
	ext         = 'pws'
	header = """
# This file is editable.
# Everything following a hash symbol '#' is a comment and ignored by {pnm}.
"""
	text_label_header = """
# A text label of {n} screen cells or less may be added to the right of each
# password.  The label may contain any printable ASCII symbol.
"""

	def get_line(self, lines):

		self.line_ctr += 1
		p = self.parent

		if p.pw_fmt in ('bip39', 'xmrseed'):
			ret = lines.pop(0).split(None, p.pw_len+1)
			if len(ret) > p.pw_len+1:
				m1 = f'extraneous text {ret[p.pw_len+1]!r} found after password'
				m2 = '[bare comments not allowed in BIP39 password files]'
				m = m1+' '+m2
			elif len(ret) < p.pw_len+1:
				m = f'invalid password length {len(ret)-1}'
			else:
				return (ret[0], ' '.join(ret[1:p.pw_len+1]), '')
			raise ValueError(m)
		else:
			ret = lines.pop(0).split(None, 2)
			return ret if len(ret) == 3 else ret + ['']

	def make_label(self):
		p = self.parent
		return f'{p.al_id.sid} {p.pw_id_str} {p.pw_fmt_disp}:{p.pw_len}'



#!/usr/bin/env python3

"""
test.modtest_d.mn_entry: Mnemonic user entry unit test for the MMGen suite
"""

from mmgen.mn_entry import mn_entry
from ..include.common import cfg, vmsg

class unit_tests:

	vectors = {
		'mmgen':   {
			'usl': 10, 'sw': 3, 'lw': 12,
			'idx_minimal': ( # None: non-unique match. False: no match
				('a',          None),
				('aa',         False),
				('as',         None),
				('ask',        70),
				('afte',       None),
				('after',      None),
				('aftern',     20),
				('afternoon',  20),
				('afternoons', False),
				('g',          None),
				('gg',         False),
				('z',          False),
				('abi',        False),
				('abo',        None),
				('abl',        0),
				('able',       0),
				('abler',      False),
				('you',        None),
				('yout',       1625),
				('youth',      1625),
				('youths',     False),
			),
		},
		'xmrseed': {'usl': 3, 'sw': 4, 'lw': 12},
		'bip39':   {'usl': 4, 'sw': 3, 'lw': 8},
	}

	def wl(self, name, ut, desc='MnemonicEntry - computed wordlist constants'):
		for wl_id in self.vectors:
			for j, k in (
					('uniq_ss_len',   'usl'),
					('shortest_word', 'sw'),
					('longest_word',  'lw')):
				a = getattr(mn_entry(cfg, wl_id), j)
				b = self.vectors[wl_id][k]
				assert a == b, f'{wl_id}:{j} {a} != {b}'
		return True

	def idx(self, name, ut, desc='MnemonicEntry - idx()'):
		junk = 'a g z aa gg zz aaa ggg zzz aaaa gggg zzzz aaaaaaaaaaaaaa gggggggggggggg zzzzzzzzzzzzzz'
		for wl_id in self.vectors:
			m = mn_entry(cfg, wl_id)
			vmsg('Wordlist: '+wl_id)
			for entry_mode in ('full', 'short'):
				for a, word in enumerate(m.wl):
					b = m.idx(word, entry_mode)
					assert a == b, f'{a} != {b} ({word!r} - entry mode: {entry_mode!r})'
				a = None
				for word in junk.split():
					b = m.idx(word, entry_mode)
					assert a == b, f'{a} != {b} ({word!r} - entry mode: {entry_mode!r})'
			if 'idx_minimal' in self.vectors[wl_id]:
				for vec in self.vectors[wl_id]['idx_minimal']:
					chk = vec[1]
					b = m.idx(vec[0], 'minimal')
					if chk is False:
						assert b is None, (b, None)
					elif chk is None:
						assert type(b) is tuple, (type(b), tuple)
					elif type(chk) is int:
						assert b == chk, (b, chk)
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
seed: Seed-related classes and methods for the MMGen suite
"""

from .util import make_chksum_8, hexdigits_uc, die
from .objmethods import HiliteStr, InitErrors, MMGenObject
from .obj import ImmutableAttr, get_obj

class SeedID(HiliteStr, InitErrors):
	color = 'blue'
	width = 8
	trunc_ok = False
	def __new__(cls, *, seed=None, sid=None):
		if isinstance(sid, cls):
			return sid
		try:
			if seed:
				assert isinstance(seed, SeedBase), 'not a subclass of SeedBase'
				return str.__new__(cls, make_chksum_8(seed.data))
			elif sid:
				assert set(sid) <= set(hexdigits_uc), 'not uppercase hex digits'
				assert len(sid) == cls.width, f'not {cls.width} characters wide'
				return str.__new__(cls, sid)
			raise ValueError('no arguments provided')
		except Exception as e:
			return cls.init_fail(e, seed or sid)

def is_seed_id(s):
	return get_obj(SeedID, sid=s, silent=True, return_bool=True)

class SeedBase(MMGenObject):

	lens = (128, 192, 256)
	dfl_len = 256

	data = ImmutableAttr(bytes, typeconv=False)
	sid  = ImmutableAttr(SeedID, typeconv=False)

	def __init__(self, cfg, *, seed_bin=None, nSubseeds=None):

		if not seed_bin:
			from .crypto import Crypto
			from hashlib import sha256
			# Truncate random data for smaller seed lengths
			seed_bin = sha256(Crypto(cfg).get_random(1033)).digest()[:(cfg.seed_len or self.dfl_len)//8]
		elif len(seed_bin)*8 not in self.lens:
			die(3, f'{len(seed_bin)*8}: invalid seed bit length')

		self.cfg = cfg
		self.data = seed_bin
		self.sid  = SeedID(seed=self)
		self.nSubseeds = nSubseeds # overrides cfg.subseeds

	@property
	def bitlen(self):
		return len(self.data) * 8

	@property
	def byte_len(self):
		return len(self.data)

	@property
	def fn_stem(self):
		return self.sid

class Seed(SeedBase):

	@property
	def subseeds(self):
		if not hasattr(self, '_subseeds'):
			from .subseed import SubSeedList
			self._subseeds = SubSeedList(
				self,
				length = self.nSubseeds or self.cfg.subseeds)
		return self._subseeds

	def subseed(self, *args, **kwargs):
		return self.subseeds.get_subseed_by_ss_idx(*args, **kwargs)

	def subseed_by_seed_id(self, *args, **kwargs):
		return self.subseeds.get_subseed_by_seed_id(*args, **kwargs)

	def split(self, *args, **kwargs):
		from .seedsplit import SeedShareList
		return SeedShareList(self, *args, **kwargs)

	@staticmethod
	def join_shares(*args, **kwargs):
		from .seedsplit import join_shares
		return join_shares(*args, **kwargs)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.txhistory: Tracking wallet transaction history class for the MMGen suite
"""

from collections import namedtuple

from ..util import fmt
from ..obj import NonNegativeInt
from .view import TwView

class TwTxHistory(TwView):

	class display_type(TwView.display_type):

		class squeezed(TwView.display_type.squeezed):
			cols = ('num', 'txid', 'date', 'inputs', 'amt', 'outputs', 'comment')
			subhdr_fmt_method = 'gen_squeezed_subheader'

		class detail(TwView.display_type.detail):
			need_column_widths = False
			subhdr_fmt_method = 'gen_detail_subheader'
			colhdr_fmt_method = None
			item_separator = '\n\n'

	has_wallet = False
	show_txid = False
	show_unconfirmed = False
	show_total_amt = False
	update_widths_on_age_toggle = True
	print_output_types = ('squeezed', 'detail')
	filters = ('show_unconfirmed',)
	mod_subpath = 'tw.txhistory'

	async def __init__(self, cfg, proto, *, sinceblock=0):
		await super().__init__(cfg, proto)
		self.sinceblock = NonNegativeInt(sinceblock if sinceblock >= 0 else self.rpc.blockcount + sinceblock)

	@property
	def no_rpcdata_errmsg(self):
		return 'No transaction history {}found!'.format(
			f'from block {self.sinceblock} ' if self.sinceblock else '')

	def filter_data(self):
		return (d for d in self.data if d.confirmations > 0 or self.show_unconfirmed)

	def set_amt_widths(self, data):
		amts_tuple = namedtuple('amts_data', ['amt'])
		return super().set_amt_widths([amts_tuple(d.amt_disp(self.show_total_amt)) for d in data])

	def get_column_widths(self, data, *, wide, interactive):

		# var cols: inputs outputs comment [txid]
		if not hasattr(self, 'varcol_maxwidths'):
			self.varcol_maxwidths = {
				'inputs': max(len(d.vouts_disp(
					'inputs', width=None, color=False, addr_view_pref=self.addr_view_pref)) for d in data),
				'outputs': max(len(d.vouts_disp(
					'outputs', width=None, color=False, addr_view_pref=self.addr_view_pref)) for d in data),
				'comment': max(len(d.comment) for d in data),
			}

		maxws = self.varcol_maxwidths.copy()
		minws = {
			'inputs': 15,
			'outputs': 15,
			'comment': len('Comment'),
		}
		if self.show_txid:
			maxws['txid'] = self.txid_w
			minws['txid'] = 8
			maxws_nice = {'txid': 20}
		else:
			maxws['txid'] = 0
			minws['txid'] = 0
			maxws_nice = {}

		widths = { # fixed cols
			'num': max(2, len(str(len(data)))+1),
			'date': self.age_w,
			'amt': self.amt_widths['amt'],
			'spc': 6 + self.show_txid, # 5(6) spaces between cols + 1 leading space in fs
		}

		return self.compute_column_widths(
			widths,
			maxws,
			minws,
			maxws_nice,
			wide        = wide,
			interactive = interactive)

	def gen_squeezed_subheader(self, cw, color):
		# keep these shorter than min screen width (currently prompt width, or 65 chars)
		if self.sinceblock:
			yield f'Displaying transactions since block {self.sinceblock.hl(color=color)}'
		yield 'Only wallet-related outputs are shown'
		yield 'Comment is from first wallet address in outputs or inputs'
		if (cw.inputs < self.varcol_maxwidths['inputs'] or
			cw.outputs < self.varcol_maxwidths['outputs']):
			yield 'Note: screen is too narrow to display all inputs and outputs'

	def gen_detail_subheader(self, cw, color):
		if self.sinceblock:
			yield f'Displaying transactions since block {self.sinceblock.hl(color=color)}'
		yield 'Only wallet-related outputs are shown'

	def squeezed_col_hdr(self, cw, fs, color):
		return fs.format(
			n = '',
			t = 'TxID',
			d = self.age_hdr,
			i = 'Inputs',
			A = 'Amt({})'.format('TX' if self.show_total_amt else 'Wallet'),
			o = 'Outputs',
			c = 'Comment')

	def gen_squeezed_display(self, data, cw, fs, color, fmt_method):

		for n, d in enumerate(data, 1):
			yield fs.format(
				n = str(n) + ')',
				t = d.txid_disp(width=cw.txid, color=color) if hasattr(cw, 'txid') else None,
				d = d.age_disp(self.age_fmt, width=self.age_w, color=color),
				i = d.vouts_disp('inputs', width=cw.inputs, color=color, addr_view_pref=self.addr_view_pref),
				A = d.amt_disp(self.show_total_amt).fmt(cw.iwidth, prec=self.disp_prec, color=color),
				o = d.vouts_disp('outputs', width=cw.outputs, color=color, addr_view_pref=self.addr_view_pref),
				c = d.comment.fmt2(cw.comment, color=color, nullrepl='-'))

	def gen_detail_display(self, data, cw, fs, color, fmt_method):

		fs = fmt("""
		{n}
		    Block:        [{d}] {b}
		    TxID:         [{D}] {t}
		    Value:        {A}
		    Wallet Value: {B}
		    Fee:          {f}
		    Inputs:
		        {i}
		    Outputs ({N}):
		        {o}
		""", strip_char='\t').strip()

		for n, d in enumerate(data, 1):
			yield fs.format(
				n = str(n) + ')',
				d = d.age_disp('date_time', width=None, color=None),
				b = d.blockheight_disp(color=color),
				D = d.txdate_disp('date_time'),
				t = d.txid_disp(color=color),
				A = d.amt_disp(show_total_amt=True).hl(color=color),
				B = d.amt_disp(show_total_amt=False).hl(color=color),
				f = d.fee_disp(color=color),
				i = d.vouts_list_disp('inputs', color=color, indent=' '*8, addr_view_pref=self.addr_view_pref),
				N = d.nOutputs,
				o = d.vouts_list_disp('outputs', color=color, indent=' '*8, addr_view_pref=self.addr_view_pref),
			)

	sort_disp = {
		'age':         'Age',
		'blockheight': 'Block Height',
		'amt':         'Wallet Amt',
		'total_amt':   'TX Amt',
		'txid':        'TxID',
	}

	sort_funcs = {
		'age':         lambda i: '{:010}.{:010}'.format(0xffffffff - abs(i.confirmations), i.time_received or 0),
		'blockheight': lambda i: 0 - abs(i.confirmations), # old/altcoin daemons return no 'blockheight' field
		'amt':         lambda i: i.wallet_outputs_total,
		'total_amt':   lambda i: i.outputs_total,
		'txid':        lambda i: i.txid,
	}

	async def set_dates(self, _):
		pass

	@property
	def dump_fn_pfx(self):
		return 'transaction-history' + (f'-since-block-{self.sinceblock}' if self.sinceblock else '')

	class sort_action(TwView.sort_action):

		def s_blockheight(self, parent):
			parent.do_sort('blockheight')

		def s_amt(self, parent):
			parent.do_sort('amt')
			parent.show_total_amt = False

		def s_total_amt(self, parent):
			parent.do_sort('total_amt')
			parent.show_total_amt = True

	class display_action(TwView.display_action):

		def d_show_txid(self, parent):
			parent.show_txid = not parent.show_txid

		def d_show_unconfirmed(self, parent):
			parent.show_unconfirmed = not parent.show_unconfirmed

		def d_show_total_amt(self, parent):
			parent.show_total_amt = not parent.show_total_amt
